package com.ringlord;

import java.io.File;


/**
 * Filename handling.
 * 
 * @author K Udo Schuermann
 */
public class Filename
{
  /**
   * <p>
   * Resolves the Unix/Linux ~ (tilde) convention for filenames, producing a
   * {@link File} from a String that may include a leading tilde.
   * </p>
   * 
   * <p>
   * The tilde represents a reference to the home directory. Without a username
   * following, the tilde resolves to the current user's home directory,
   * otherwise the name following the tilde references that user's home
   * directory, instead. Examples:
   * </p>
   * 
   * <ul>
   * <li><tt>~/myfile</tt> is <tt>/home/udo/myfile</tt> (for current user 'udo')
   * <li><tt>~/myfile</tt> is <tt>/home/jhendrix/myfile</tt> (for current user
   * 'jhendrix')
   * <li><tt>~udo/myfile</tt> is <tt>/home/udo/myfile</tt> (regardless of
   * current user)
   * <li><tt>~jhendrix/myfile</tt> is <tt>/home/jhendrix/myfile</tt> (regardless
   * of current user)
   * </ul>
   * 
   * <p>
   * Assumes that all user directories are under the same root. In other words,
   * if the current user's home directory is <tt>/home/jhendrix</tt>, then user
   * udo's home directory will be <tt>/home/udo</tt>
   * </p>
   * 
   * <p>
   * Note that Microsoftian '\' characters are not recognized for the tilde
   * construct. In other words, <tt>~jhendrix\myfile</tt> and <tt>~\myfile</tt>
   * will not work.
   * </p>
   * 
   * @param filename
   * @return
   */
  public static File resolveTilde( String filename )
  {
    if( filename.startsWith( "~" ) )
      {
	if( filename.startsWith( "~/" ) )
	  {
	    filename = System.getProperty( "user.home" ) + filename.substring( 1 );
	  }
	else
	  // it's something like "~userid/..."
	  {
	    final int slashIndex = filename.indexOf( "/" );
	    final String userID = filename.substring( 1,
		                                      slashIndex );
	    final File userHome = new File( System.getProperty( "user.home" ) );
	    final File homes = userHome.getParentFile();
	    final File base = new File( homes,
		                        userID );
	    filename = new File( base,
		                 filename.substring( slashIndex + 1 ) ).toString();
	  }
      }
    return new File( filename );
  }

}
