package com.ringlord;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


/**
 * Provides a mechanism to invoke an external program with parameters, supply it
 * stdin, and capture stdout and stderr without the risk of blocking and hanging
 * the process. The {@link #execute} methods will block until the process exits
 * on its own, or is {@link #destroy() destroyed} prematurely.
 * 
 * @author K. Udo Schuermann
 **/
public class Runner
{
  /**
   * Construct a Runner.
   */
  public Runner()
  {
    super();
  }


  /**
   * Notify a Runner to terminate a process that is currently {@link #execute
   * executing}.
   */
  public void destroy()
  {
    synchronized( this )
      {
	notify();
      }
  }


  public void execute( final String ... commandline )
    throws IOException
  {
    execute( null,
	     null,
	     null,
	     commandline );
  }


  /**
   * Using input and output streams (any of which are optional and can be null),
   * invoke a given program and return only when the program has terminated.
   **/
  public synchronized void execute( final InputStream stdin,
	                            final OutputStream stdout,
	                            final OutputStream stderr,
	                            final String ... commandline )
    throws IOException
  {
    final Process p = Runtime.getRuntime().exec( commandline );
    final Runner me = this;
    final Thread t = new Thread()
    {
      @Override
      public void run()
      {
	Firehose out = null;
	Firehose err = null;
	Firehose in = null;

	try
	  {
	    if( stdout != null )
	      {
		final InputStream inputStream = p.getInputStream();
		if( inputStream != null )
		  {
		    out = new Firehose( inputStream,
			                stdout,
			                "stdout: " + commandline[0],
			                512 );
		    out.start();
		  }
	      }

	    if( stderr != null )
	      {
		final InputStream errorStream = p.getErrorStream();
		if( errorStream != null )
		  {
		    err = new Firehose( errorStream,
			                stderr,
			                "stderr: " + commandline[0],
			                512 );
		    err.start();
		  }
	      }

	    if( stdin != null )
	      {
		final OutputStream outputStream = p.getOutputStream();
		if( outputStream != null )
		  {
		    in = new Firehose( stdin,
			               outputStream,
			               "stdin: " + commandline[0],
			               512 );
		    in.start();
		    try
		      {
			in.join();
		      }
		    catch( final InterruptedException x )
		      {
		      }
		    try
		      {
			outputStream.close();
		      }
		    catch( final IOException x )
		      {
		      }
		  }
	      }

	    p.waitFor();

	    if( out != null )
	      {
		out.join();
	      }

	    if( err != null )
	      {
		err.join();
	      }
	  }
	catch( final InterruptedException x )
	  {
	    // destroy() was called to cancel execution
	  }
	finally
	  {
	    if( in != null )
	      {
		in.die();
	      }
	    if( out != null )
	      {
		out.die();
	      }
	    if( err != null )
	      {
		err.die();
	      }
	    p.destroy(); // kill the process
	    synchronized( me )
	      {
		// stop the Runner
		me.notify();
	      }
	  }
      }
    };

    synchronized( this )
      {
	try
	  {
	    t.start();
	    wait();
	  }
	catch( final InterruptedException x )
	  {
	  }
	t.interrupt();
      }
  }


  private static class Firehose
    extends Thread
  {
    Firehose( final InputStream in,
	      final OutputStream out,
	      final String name,
	      final int bufferSize )
    {
      super( name );
      this.in = in;
      this.out = out;
      this.bufferSize = bufferSize;
    }


    void die()
    {
      interrupt();
    }


    @Override
    public void run()
    {
      final byte[] buffer = new byte[bufferSize];
      try
	{
	  int inBuffer;
	  if( out == null )
	    {
	      while( (inBuffer = in.read( buffer )) > -1 )
		{
		  // no destination, just eat all the input
		}
	    }
	  else
	    {
	      while( (inBuffer = in.read( buffer )) > -1 )
		{
		  out.write( buffer,
			     0,
			     inBuffer );
		}
	      out.flush();
	    }
	}
      catch( final Throwable x )
	{
	}
    }
    private final InputStream in;
    private final OutputStream out;
    private final int bufferSize;
  }
}
