package com.ringlord;

import java.awt.Rectangle;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.prefs.Preferences;

import com.ringlord.mime.Base64;


/**
 * <p>
 * A simple wrapper around standard {@link Preferences}. The following
 * demonstrates ways to use this class:
 * </p>
 *
 * <dl>
 * <dt><code>Settings config = getInstance();</code>
 * <dd>Obtain the the Default Settings. The first time this method is called,
 * the caller is considered the owner and the Settings are constructed for that
 * package.
 * <dt><code>Settings config = getInstance( getClass() );</code>
 * <dd>Obtain the Settings for a specifid class (package). If there are no
 * Default Settings defined, then this call will register the returned Settings
 * as the Default Settings.
 * <dt><code>String foo = config.get( "my-value", (String)null );</code>
 * <dd>Obtain a String setting identified by "my-value", and return null if it
 * has no value.
 * <dt><code>boolean bar = config.get( "light-on", true );</code> Obtain a
 * boolean value identified by "light-on"; if it has no value yet, then return
 * true (otherwise the default is ignored).
 * <dt><code>List&lt;String&gt; names = get( "names", null, "{sep}" );</code>
 * <dd>Obtain a list of Strings identified by "names"; no default is returned if
 * the list is not defined; items in the list are separated with the
 * multi-character sequence "{sep}".
 * </dl>
 *
 * @author K. Udo Schuermann
 * @see Config
 **/
public class Settings
{
  public static synchronized Settings getInstance()
    throws ClassNotFoundException
  {
    if( defaultSettings == null )
      {
        final Throwable x = new Throwable( "" );
        final StackTraceElement[] trace = x.getStackTrace();
        final Class<?> c = Class.forName( trace[1].getClassName() );
        getInstance( c ); // side-effect: sets defaultSettings if null
      }
    return defaultSettings;
  }


  public static synchronized Settings getInstance( final Class<?> owner )
  {
    if( owner == null )
      {
        throw new IllegalArgumentException( "Requires a non-null Class" );
      }
    Settings instance = instances.get( owner );
    if( instance == null )
      {
        instance = new Settings( owner );
        instances.put( owner,
                       instance );
        if( defaultSettings == null )
          {
            defaultSettings = instance;
          }
      }
    return instance;
  }


  public static void setDefaultSettings( final Class<?> owner )
  {
    Settings instance = instances.get( owner );
    if( instance == null )
      {
        instance = new Settings( owner );
        instances.put( owner,
                       instance );
      }
    defaultSettings = instance;
  }


  public void put( final String key,
                   final List<String> value,
                   final String separator )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        final StringBuilder sb = new StringBuilder();
        boolean isFirst = true;
        if( separator != null )
          {
            for( final String s : value )
              {
                if( isFirst )
                  {
                    isFirst = false;
                  }
                else
                  {
                    sb.append( separator );
                  }
                sb.append( s );
              }
          }
        else
          {
            for( final String s : value )
              {
                if( isFirst )
                  {
                    isFirst = false;
                  }
                else
                  {
                    sb.append( "," );
                  }
                final byte[] bytes = Base64.encode( s.getBytes( utf8 ) );
                sb.append( new String( bytes ) );
              }
          }
        prefs.put( key,
                   sb.toString() );
      }
  }


  /**
   * Obtain a List of (String) items.
   *
   * @param key
   *          The key for the list
   * @param defaultValue
   *          A default List if the setting is undefined
   * @param separator
   *          An optional separator: If no separator is defined (null or
   *          zero-length), the items are considered Base64 encoded and
   *          separated with commas, otherwise the separator (which may be
   *          anything, including newline or even multiple characters) is used
   *          to break the value into multiple items.
   * @return
   */
  public List<String> get( final String key,
                           final List<String> defaultValue,
                           final String separator )
  {
    final String tmp = prefs.get( key,
                                  (String)null );
    if( tmp == null )
      {
        return defaultValue;
      }
    final List<String> result = new ArrayList<>();
    int start = 0;
    int pos;
    if( (separator != null) && (separator.length() > 0) )
      {
        final int len = separator.length();
        while( (pos = tmp.indexOf( separator,
                                   start )) > -1 )
          {
            result.add( tmp.substring( start,
                                       pos ) );
            start = pos + len;
          }
        result.add( tmp.substring( start ) );
      }
    else
      {
        while( (pos = tmp.indexOf( ",",
                                   start )) > -1 )
          {
            final byte[] bytes = Base64.decode( tmp.substring( start,
                                                               pos ).getBytes() );
            result.add( new String( bytes,
                                    utf8 ) );
            start = pos + 1;
          }
        final byte[] bytes = Base64.decode( tmp.substring( start ).getBytes() );
        result.add( new String( bytes,
                                utf8 ) );
      }
    return result;
  }


  // ======================================================================

  public void put( final String key,
                   final String value )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.put( key,
                   value );
      }
  }


  public String get( final String key,
                     final String defaultValue )
  {
    return prefs.get( key,
                      defaultValue );
  }


  // ======================================================================
  public void put( final String key,
                   final Rectangle rect )
  {
    if( rect == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.put( key,
                   rect.x + "," + rect.y + "," + rect.width + "," + rect.height );
      }
  }


  public Rectangle get( final String key,
                        final Rectangle defaultValue )
  {
    final String value = prefs.get( key,
                                    null );
    if( value == null )
      {
        return defaultValue;
      }

    try
      {
        int prev, comma;

        comma = value.indexOf( ',' );
        final int x = Integer.parseInt( value.substring( 0,
                                                         comma ) );

        prev = comma + 1;
        comma = value.indexOf( ',',
                               prev );
        final int y = Integer.parseInt( value.substring( prev,
                                                         comma ) );

        prev = comma + 1;
        comma = value.indexOf( ',',
                               prev );
        final int w = Integer.parseInt( value.substring( prev,
                                                         comma ) );

        prev = comma + 1;
        final int h = Integer.parseInt( value.substring( prev ) );

        return new Rectangle( x,
                              y,
                              w,
                              h );
      }
    catch( final Exception x )
      {
        x.printStackTrace();
        return defaultValue;
      }
  }


  // ======================================================================
  public void put( final String key,
                   final Boolean value )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.putBoolean( key,
                          value );
      }
  }


  public boolean get( final String key,
                      final boolean defaultValue )
  {
    return prefs.getBoolean( key,
                             defaultValue );
  }


  // ======================================================================
  public void put( final String key,
                   final Double value )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.putDouble( key,
                         value );
      }
  }


  public double get( final String key,
                     final double defaultValue )
  {
    return prefs.getDouble( key,
                            defaultValue );
  }


  // ======================================================================

  public void put( final String key,
                   final Long value )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.putLong( key,
                       value );
      }
  }


  public long get( final String key,
                   final long defaultValue )
  {
    return prefs.getLong( key,
                          defaultValue );
  }


  // ======================================================================

  public void put( final String key,
                   final Integer value )
  {
    if( value == null )
      {
        prefs.remove( key );
      }
    else
      {
        prefs.putInt( key,
                      value );
      }
  }


  public int get( final String key,
                  final int defaultValue )
  {
    return prefs.getInt( key,
                         defaultValue );
  }


  // Hide the constructor
  private Settings( final Class<?> owner )
  {
    super();
    this.prefs = Preferences.userNodeForPackage( owner );
  }

  private final Preferences prefs;
  //
  private static Settings defaultSettings;
  private static final Map<Class<?>,Settings> instances = new HashMap<>();
  private static final Charset utf8 = Charset.forName( "UTF-8" );
}
