package com.ringlord.boxer2;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipException;

import com.ringlord.Filename;
import com.ringlord.Runner;
import com.ringlord.boxer2.Manifest.AddResult;
import com.ringlord.boxer2.Project.Addition;
import com.ringlord.boxer2.Project.Addition.Where;
import com.ringlord.boxer2.Project.Classpath;
import com.ringlord.boxer2.Project.Library;
import com.ringlord.boxer2.Project.ProjectReference;
import com.ringlord.boxer2.Project.Sources;
import com.ringlord.boxer2.SigningProperties.PasswordReader;


/**
 * Static methods to produce Boxer's output products (JAR file and distributable
 * archive).
 *
 * @author K Udo Schuermann
 */
public class Builder
{
  public static void writeJAR( final Project project,
                               final File jarFile,
                               final Manifest manifest )
    throws FileNotFoundException,
      IOException
  {
    final long startTime = System.currentTimeMillis();

    final StringBuilder jarManifest = new StringBuilder();
    jarManifest.append( toJarLine( "Manifest-Version: 1.0" ) );
    jarManifest.append( toJarLine( "Created-By: " + System.getProperty( "java.version" ) ) );
    for( String line : project.getManifestAdditions() )
      {
        line = line.trim();
        final int colonIndex = line.indexOf( ":" );
        if( (colonIndex > 0) && (line.length() > 2) )
          {
            // We have at LEAST "k:v"
            final String key = line.substring( 0,
                                               colonIndex ).trim();
            final String val = Util.resolve( project,
                                             line.substring( colonIndex + 1 ) ).trim();

            if( "Manifest-Version".equalsIgnoreCase( key ) ||
                "Created-By".equalsIgnoreCase( key ) ||
                "Producer".equalsIgnoreCase( key ) ||
                key.toUpperCase().startsWith( "BOXER-" ) )
              {
                Util.logger().log( Level.WARNING,
                                   "Ignoring extraneous JAR manifest addition: " + line );
              }
            else
              {
                jarManifest.append( toJarLine( key + ": " + val ) );
              }
          }
        else
          {
            Util.logger().log( Level.WARNING,
                               "Ignoring malformed JAR manifest addition: " + line );
          }
      }
    jarManifest.append( toJarLine( "Producer: Boxer " + Util.VERSION + " (" + Util.VERDATE + ")" ) );
    jarManifest.append( toJarLine( "Boxer-Vendor: Ringlord Technologies" ) );
    jarManifest.append( toJarLine( "Boxer-Homepage: http://ringlord.com/boxer/" ) );

    final String mainClass = manifest.getMainClassName();
    if( mainClass != null )
      {
        jarManifest.append( toJarLine( "Main-Class: " + mainClass ) );
      }

    final String splashImageName = project.getSplashImageName();
    if( (splashImageName != null) && (splashImageName.length() > 0) )
      {
        jarManifest.append( toJarLine( "SplashScreen-Image: " + (splashImageName.startsWith( "/" )
            ? ""
            : "/") + splashImageName ) );
      }
    final List<String> classpath = manifest.getClassPathEntries();
    if( !classpath.isEmpty() )
      {
        boolean isFirst = true;
        final StringBuilder cp = new StringBuilder();
        for( final String s : classpath )
          {
            if( isFirst )
              {
                isFirst = false;
              }
            else
              {
                cp.append( ' ' );
              }
            cp.append( s );
          }
        jarManifest.append( toJarLine( "Class-Path: " + cp.toString() ) );
      }

    final byte[] m = jarManifest.toString().getBytes();

    try( final PackageWriterZip jar = new PackageWriterZip( jarFile ) )
      {
        final ByteArrayInputStream manifestStream = new ByteArrayInputStream( m );
        jar.add( manifestStream,
                 "META-INF/MANIFEST.MF",
                 System.currentTimeMillis() );

        for( final String outputName : manifest )
          {
            final File sourceFile = manifest.getFile( outputName );
            try
              {
                jar.add( sourceFile,
                         outputName );
              }
            catch( final ZipException x )
              {
                Util.logger().log( Level.SEVERE,
                                   "Failed to store " + sourceFile,
                                   x );
              }
          }
      }

    final long totalTime = System.currentTimeMillis() - startTime;
    Util.logger().log( Level.INFO,
                       String.format( "Finished building %s in %1.3f seconds.",
                                      jarFile.toString(),
                                      totalTime / 1000.0d ) );
  }


  public static void signJAR( final Project project,
                              final File jarFile,
                              final PasswordReader passwordReader )
  {
    // keystore information, especially the store and key passwords, are
    // kept in a file that should be external to the project.
    if( project.getSignJAR() )
      {
        if( getJarSignerBinary() != null )
          {
            final String signingPropertiesFilename = project.getSigningPropertiesFile();
            if( (signingPropertiesFilename != null) && (signingPropertiesFilename.length() > 0) )
              {
                final File signingPropertiesFile = Filename.resolveTilde( signingPropertiesFilename );

                try
                  {
                    final long startTimeGetSecrets = System.currentTimeMillis();
                    final SigningProperties secrets = SigningProperties.load( signingPropertiesFile,
                                                                              passwordReader );
                    if( secrets == null )
                      {
                        Util.logger().log( Level.WARNING,
                                           "Cannot sign JAR: Did not obtain secrets to access "
                                               + "the KeyStore file for signing the JAR." );
                        return;
                      }
                    final long startTimeSignJAR = System.currentTimeMillis();

                    final String storeFilename = secrets.getKeystoreFilename();
                    final String storePassword = secrets.getKeystorePassword();
                    final String keyAlias = secrets.getKeyAlias();
                    final String keyPassword = secrets.getKeyPassword();

                    // Feed passwords through stdin so that they will not appear
                    // the process listing in any way, shape, or form, even for
                    // just one moment.
                    final byte[] jarSignerInput = (storePassword + "\n" + keyPassword + "\n").getBytes();

                    final ByteArrayInputStream in = new ByteArrayInputStream( jarSignerInput );
                    final ByteArrayOutputStream out = new ByteArrayOutputStream();
                    final ByteArrayOutputStream err = new ByteArrayOutputStream();

                    final Runner runner = new Runner();
                    runner.execute( in,
                                    out,
                                    err,
                                    "jarsigner",
                                    "-keystore",
                                    storeFilename,
                                    jarFile.toString(),
                                    keyAlias );

                    final long endTimeSignJAR = System.currentTimeMillis();

                    final Logger logger = Util.logger();
                    final String info1 = String.format( "Time to retrieve secrets: %1.3f seconds.",
                                                        (startTimeSignJAR - startTimeGetSecrets) / 1000.0d );
                    final String info2 = String.format( "Finished signing JAR file in %1.3f seconds.",
                                                        (endTimeSignJAR - startTimeSignJAR) / 1000.0d );
                    logger.log( Level.INFO,
                                info1 );

                    final String stdout = out.toString().trim();
                    if( stdout.length() > 0 )
                      {
                        logger.log( Level.INFO,
                                    "JarSigner produced the following output:\n" + stdout );
                      }

                    logger.log( Level.INFO,
                                info2 );
                  }
                catch( IllegalStateException |
                       IOException x )
                  {
                    Util.logger().log( Level.SEVERE,
                                       "Cannot sign JAR (" + jarFile + ")",
                                       x );
                  }
              }
            else
              {
                Util.logger().log( Level.WARNING,
                                   "Cannot sign JAR: No signing properties file specified "
                                       + "from which to obtain secrets to access the "
                                       + "KeyStore file for signing the JAR." );
              }
          }
        else
          {
            Util.logger().log( Level.SEVERE,
                               "Cannot sign JAR: jarsigner binary is not available" );
          }
      }
  }


  public static void writeJNLP( final Project project,
                                final File archive )
  {
    if( project.getBuildJNLP() )
      {
        final long startTime = System.currentTimeMillis();
        final String jnlp = project.getJNLP();
        if( jnlp.trim().length() > 0 )
          {
            String name = archive.toString();
            final int lastDot = name.lastIndexOf( "." );
            if( lastDot > -1 )
              {
                name = name.substring( 0,
                                       lastDot );
              }
            name = name + ".jnlp";

            String encoding = "UTF-8";
            try
              {
                final Matcher m = Pattern.compile( "encoding\\s*=\\s*\".*?\"" ).matcher( jnlp );
                if( m.find() )
                  {
                    encoding = m.group( 1 );
                  }
              }
            catch( final Exception x )
              {
              }

            try( final OutputStream jnlpFile = new FileOutputStream( name ) )
              {
                try( final OutputStreamWriter f = new OutputStreamWriter( jnlpFile,
                                                                          Charset.forName( encoding ) ) )
                  {
                    f.write( jnlp );
                    Util.logger().log( Level.INFO,
                                       String.format( "Finished building %s in %1.3f seconds.",
                                                      name,
                                                      (System.currentTimeMillis() - startTime) / 1000.0d ) );
                    return;
                  }
              }
            catch( final IOException x )
              {
                Util.logger().log( Level.SEVERE,
                                   "Failed to write JNLP file: " + name,
                                   x );
              }
          }

        Util.logger().log( Level.SEVERE,
                           "Not yet able to build JNLP file" );
      }
  }


  public static void writeDistro( final File archive,
                                  final File jar,
                                  final String subdir,
                                  final List<Project.Library> libraries,
                                  final List<Addition> additions )
    throws IOException
  {
    final long startTime = System.currentTimeMillis();

    final String ucArchiveName = archive.toString().toUpperCase();
    if( ucArchiveName.endsWith( ".ZIP" ) )
      {
        try( final PackageWriterZip zip = new PackageWriterZip( archive ) )
          {
            writeDistro( zip,
                         jar,
                         subdir,
                         libraries,
                         additions );
          }
      }
    else if( ucArchiveName.endsWith( ".TAR" ) || ucArchiveName.endsWith( ".TAR.GZ" ) || ucArchiveName.endsWith( ".TGZ" ) )
      {
        // Write to a temporary file (same name, but .tmp extension) first
        final File outFile;
        if( ucArchiveName.endsWith( ".TAR.GZ" ) )
          {
            outFile = new File( archive.toString().substring( 0,
                                                              ucArchiveName.length() - 7 ) + ".tmp" );
          }
        else if( ucArchiveName.endsWith( ".TGZ" ) )
          {
            outFile = new File( archive.toString().substring( 0,
                                                              ucArchiveName.length() - 4 ) + ".tmp" );
          }
        else if( ucArchiveName.endsWith( ".TAR" ) )
          {
            outFile = archive;
          }
        else
          {
            Util.logger().log( Level.SEVERE,
                               "Internal failure: Not all filename extension cases were handled" );
            return;
          }

        try( final PackageWriterTar tar = new PackageWriterTar( outFile ) )
          {
            writeDistro( tar,
                         jar,
                         subdir,
                         libraries,
                         additions );
          }

        if( outFile != archive )
          {
            // We wrote to a file OTHER than the desired one, i.e. the temporary
            // one, so now it's time to compress the temporary file into the
            // actual distribution archive
            try( final GZIPOutputStream out = new GZIPOutputStream( new FileOutputStream( archive ) );
                 final FileInputStream in = new FileInputStream( outFile ) )
              {
                final byte[] BUFFER = new byte[4096];
                int inBuffer;
                while( (inBuffer = in.read( BUFFER )) > -1 )
                  {
                    out.write( BUFFER,
                               0,
                               inBuffer );
                  }
                out.finish();
              }
            // And finally, remove the temporary file
            outFile.delete();
          }
      }
    else
      {
        Util.logger().log( Level.SEVERE,
                           "Unable to create distribution archive: Filename " +
                               "extension is not among the recognized formats " +
                               "(.zip .tar .tar.gz .tgz): " +
                               archive );
        return;
      }

    final long totalTime = System.currentTimeMillis() - startTime;
    Util.logger().log( Level.INFO,
                       String.format( "Finished building %s in %1.3f seconds.",
                                      archive.toString(),
                                      totalTime / 1000.0d ) );
  }


  public static void writeDistro( final PackageWriter distro,
                                  final File jar,
                                  final String subdir,
                                  final List<Library> libraries,
                                  final List<Addition> additions )
    throws IOException
  {
    // The JAR must be added with an explicit subdirectory path
    distro.add( jar,
                new File( subdir,
                          jar.getName().toString() ).toString() );

    if( libraries != null )
      {
        for( final Project.Library library : libraries )
          {
            if( library.getHandling() == LibraryHandlingOption.REFERENCE )
              {
                final String name;
                final String refPath = library.getReferencePath();
                if( (refPath == null) || (refPath.length() == 0) )
                  {
                    name = library.archive().toString();
                  }
                else
                  {
                    name = new File( new File( subdir,
                                               refPath ),
                                     library.archive().getName() ).toString();
                  }
                distro.add( library.archive(),
                            name );
              }
          }
      }

    if( additions != null )
      {
        for( final Addition addition : additions )
          {
            final Where where = addition.where();
            if( (where == Where.BOTH) || (where == Where.ARCHIVE) )
              {
                try
                  {
                    distro.add( addition.getActualFile(),
                                new File( subdir,
                                          addition.storeAs() ).toString() );
                  }
                catch( final IOException x )
                  {
                    Util.logger().log( Level.SEVERE,
                                       "Cannot add '" +
                                           addition.getActualFile() +
                                           "' as '" +
                                           new File( subdir,
                                                     addition.storeAs() ) +
                                           "'",
                                       x );
                  }
              }
          }
      }
  }


  public static void writeDigestFiles( final Project project,
                                       final File ... files )
  {
    if( files != null )
      {
        final Map<MessageDigest,String> extensions = new HashMap<>();
        final List<MessageDigest> digests = new ArrayList<>();
        for( final String algorithm : project.getDigestsPicked() )
          {
            try
              {
                final MessageDigest digest = MessageDigest.getInstance( algorithm );
                digests.add( digest );
                final String ext = "." + algorithm.replace( "-",
                                                            "" ).toLowerCase();
                extensions.put( digest,
                                ext );
              }
            catch( final NoSuchAlgorithmException x )
              {
                Util.logger().log( Level.SEVERE,
                                   "No such algorithm available for creating digest file: " + algorithm,
                                   x );
              }
          }

        if( !digests.isEmpty() )
          {
            final byte[] BUFFER = new byte[4096];
            for( final File file : files )
              {
                if( file != null )
                  {
                    boolean isOK = true;
                    try( final InputStream f = new BufferedInputStream( new FileInputStream( file ) ) )
                      {
                        int inBuffer;
                        while( (inBuffer = f.read( BUFFER )) > -1 )
                          {
                            for( final MessageDigest digest : digests )
                              {
                                digest.update( BUFFER,
                                               0,
                                               inBuffer );
                              }
                          }
                      }
                    catch( final IOException x )
                      {
                        isOK = false;
                        Util.logger().log( Level.SEVERE,
                                           "Failed to read file to create digest file: " + file,
                                           x );
                      }
                    if( isOK )
                      {
                        String name = file.toString();
                        if( name.toUpperCase().endsWith( ".TAR.GZ" ) )
                          {
                            // special case, we need to strip both extensions
                            name = name.substring( 0,
                                                   name.length() - 7 );
                          }
                        else
                          {
                            final int lastDot = name.lastIndexOf( "." );
                            if( lastDot > -1 )
                              {
                                name = name.substring( 0,
                                                       lastDot );
                              }
                          }

                        for( final MessageDigest digest : digests )
                          {
                            final byte[] hash = digest.digest();
                            final String digestFilename = name + extensions.get( digest );
                            try( final OutputStream f = new BufferedOutputStream( new FileOutputStream( digestFilename ) ) )
                              {
                                final StringBuilder sb = new StringBuilder();
                                for( final byte b : hash )
                                  {
                                    sb.append( String.format( "%02x",
                                                              b ) );
                                  }
                                sb.append( " *" ).append( file.toString() ).append( "\n" );
                                f.write( sb.toString().getBytes() );
                                f.flush();
                                Util.logger().log( Level.CONFIG,
                                                   "Wrote digest file " + digestFilename );
                              }
                            catch( final IOException x )
                              {
                                Util.logger().log( Level.SEVERE,
                                                   "Cannot write digest file: " + digestFilename );
                              }
                            digest.reset(); // prepare for next file (if any)
                          }
                      }
                  }
              }
          }
      }
  }


  /**
   *
   * @param messages
   * @param workingPath
   * @param archive
   * @param project
   * @param exclusions
   * @param additions
   * @return
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  public static Manifest assemble( final Project project,
                                   final File workingPath,
                                   final List<Project.Library> externalLibs )
    throws NoSuchAlgorithmException,
      IOException
  {
    final Logger logger = Util.logger();
    logger.log( Level.INFO,
                "Collecting content for project " + project.name() );
    final long startTime = System.currentTimeMillis();
    workingPath.mkdirs();
    final List<Project> done = new ArrayList<>();
    final Manifest manifest = new Manifest();
    manifest.setMainClassName( project.getEntryPointName() );

    final List<Exclusion> projectExclusions = allExclusions( project );
    for( final Exclusion exclusion : projectExclusions )
      {
        logger.log( Level.FINE,
                    "Will exclude: " + exclusion.getPattern() );
      }

    final String sourceRoot = project.getSourceRoot();

    // How does the top-level project want its project reference to be handled?
    final Map<ProjectReference,ArchiveHandlingOption> projectReferenceHandling = new HashMap<>();
    final List<ProjectReference> projectReferences = project.getProjects();
    if( projectReferences != null )
      {
        for( final ProjectReference projectRef : projectReferences )
          {
            projectReferenceHandling.put( projectRef,
                                          projectRef.getHandling() );
          }
      }
    // Also capture how we wanted the project dependencies of dependent projects
    // to be handled. This is likely a very rare case, but we want to handle it.
    final List<ProjectReference> subProjectRefs = project.getSubProjects();
    if( subProjectRefs != null )
      {
        for( final ProjectReference projectRef : subProjectRefs )
          {
            projectReferenceHandling.put( projectRef,
                                          projectRef.getHandling() );
          }
      }

    final boolean[] warnBinaryExclusion = {false};
    assemble( manifest,
              workingPath,
              project,
              projectExclusions,
              sourceRoot,
              null,
              projectReferenceHandling,
              externalLibs,
              warnBinaryExclusion,
              done );

    if( warnBinaryExclusion[0] )
      {
        logger.log( Level.WARNING,
                    "Excluding binary product(s) may trigger runtime failures!" );
      }

    final long totalTime = System.currentTimeMillis() - startTime;
    Util.logger().log( Level.INFO,
                       String.format( "Collected %,d files (%,d bytes) for project %s in %1.3f seconds",
                                      manifest.fileCount(),
                                      manifest.byteCount(),
                                      project.name(),
                                      totalTime / 1000.0d ) );

    return manifest;
  }


  private static void assemble( final Manifest manifest,
                                final File workingPath,
                                final Project project,
                                final List<Exclusion> exclusions,
                                final String sourceRoot,
                                final FileHandlingOption sourceHandling,
                                final Map<ProjectReference,ArchiveHandlingOption> projectReferenceHandling,
                                final List<Project.Library> externalLibs,
                                final boolean[] warnBinaryExclusion,
                                final List<Project> done )
    throws NoSuchAlgorithmException,
      IOException
  {
    final Logger logger = Util.logger();
    logger.log( Level.INFO,
                "Processing project " + project );
    try
      {
        final List<Project.Binaries> projectBinaries = project.getBinaries();
        if( projectBinaries != null )
          {
            for( final Project.Binaries binaries : projectBinaries )
              {
                logger.log( Level.FINER,
                            "Processing " + binaries );
                for( final File file : binaries )
                  {
                    final String relativeName = binaries.toRelativeName( file );
                    if( isExcluded( relativeName,
                                    exclusions ) )
                      {
                        warnBinaryExclusion[0] = true;
                        logger.log( Level.CONFIG,
                                    "Excluding binary: " + relativeName );
                      }
                    else
                      {
                        try
                          {
                            final AddResult addResult = manifest.add( relativeName,
                                                                      file );
                            reportAddResult( addResult,
                                             logger,
                                             project,
                                             relativeName );
                          }
                        catch( final IOException x )
                          {
                            logger.log( Level.SEVERE,
                                        "Cannot add '" + file + "' as '" + relativeName + "'",
                                        x );
                          }
                      }
                  }
              }
          }
        else
          {
            logger.log( Level.FINER,
                        "No binaries" );
          }

        final List<Project.Classpath> projectClasspath = project.getClasspath();
        if( projectClasspath != null )
          {
            for( final Project.Classpath classpath : projectClasspath )
              {
                logger.log( Level.FINER,
                            "Processing " + classpath );
                for( final File file : classpath )
                  {
                    final String relativeName = classpath.toRelativeName( file );
                    if( isExcluded( relativeName,
                                    exclusions ) )
                      {
                        warnBinaryExclusion[0] = true;
                        logger.log( Level.CONFIG,
                                    "Excluding classpath entry: " + relativeName );
                      }
                    else
                      {
                        try
                          {
                            final AddResult addResult = manifest.add( relativeName,
                                                                      file );
                            reportAddResult( addResult,
                                             logger,
                                             project,
                                             relativeName );
                          }
                        catch( final IOException x )
                          {
                            logger.log( Level.SEVERE,
                                        "Cannot add '" + file + "' as '" + relativeName + "'",
                                        x );
                          }
                      }
                  }
              }
          }
        else
          {
            logger.log( Level.FINER,
                        "No classes" );
          }

        final List<Project.Library> projectLibraries = project.getLibraries();
        if( projectLibraries != null )
          {
            for( final Project.Library library : projectLibraries )
              {
                switch( library.getHandling() )
                  {
                  case IGNORE:
                    {
                      logger.log( Level.FINER,
                                  "Ignoring " + library );
                      warnBinaryExclusion[0] = true;
                      logger.log( Level.CONFIG,
                                  "Excluding library: " + library.archive() );
                    }
                    break;

                  case REFERENCE:
                    {
                      logger.log( Level.FINER,
                                  "Referencing " + library );
                      if( isExcluded( library.archive().toString(),
                                      exclusions ) )
                        {
                          warnBinaryExclusion[0] = true;
                          logger.log( Level.CONFIG,
                                      "Excluding library: " + library.archive() );
                        }
                      else
                        {
                          final String refPath = library.getReferencePath().trim();
                          final String name;
                          if( (refPath == null) || (refPath.length() == 0) )
                            {
                              name = library.archive().getName();
                            }
                          else
                            {
                              name = new File( refPath,
                                               library.archive().getName() ).toString();
                            }

                          logger.log( Level.CONFIG,
                                      "Project " + project + " references external library " + name );
                          manifest.add( name );
                          externalLibs.add( library );
                        }
                    }
                    break;

                  case MERGE:
                    {
                      logger.log( Level.FINER,
                                  "Merging " + library );
                      // Include in the manifest each file from the library
                      for( final String name : library )
                        {
                          if( isExcluded( name,
                                          exclusions ) )
                            {
                              warnBinaryExclusion[0] = true;
                              logger.log( Level.CONFIG,
                                          "Excluding library entry: " + name );
                            }
                          else
                            {
                              if( !"META-INF/MANIFEST.MF".equals( name ) &&
                                  !name.endsWith( ".SF" ) &&
                                  !name.endsWith( ".DSA" ) )
                                {
                                  final AddResult addResult = manifest.add( workingPath,
                                                                            name,
                                                                            library );
                                  reportAddResult( addResult,
                                                   logger,
                                                   project,
                                                   name );
                                }
                              else
                                {
                                  Util.logger().log( Level.FINE,
                                                     "Skipping " + name + " from " + library );
                                }
                            }
                        }
                    }
                    break;

                  default:
                    System.err.println( "LibraryHandlingOption " +
                                        library.getHandling() +
                                        " not handled in Util.assemble" );
                  }
              }
          }
        else
          {
            logger.log( Level.FINER,
                        "No libraries" );
          }

        final List<Project.Sources> projectSources = project.getSources();
        if( projectSources != null )
          {
            for( final Project.Sources source : projectSources )
              {
                switch( source.getHandling() )
                  {
                  case IGNORE:
                    logger.log( Level.FINER,
                                "Ignoring " + source );
                    break;

                  case INCLUDE:
                    FileHandlingOption srcHandling = sourceHandling;
                    if( srcHandling == null )
                      {
                        srcHandling = source.getHandling();
                      }
                    if( srcHandling == FileHandlingOption.INCLUDE )
                      {
                        logger.log( Level.FINER,
                                    "Including " + source );
                        for( final File file : source )
                          {
                            final String relativeName = source.toRelativeName( file );
                            if( isExcluded( relativeName,
                                            exclusions ) )
                              {
                                logger.log( Level.CONFIG,
                                            "Excluding source: " + relativeName );
                              }
                            else
                              {
                                final String name = ((sourceRoot == null) || (sourceRoot.length() == 0)
                                    ? relativeName
                                    : new File( sourceRoot,
                                                relativeName ).toString());
                                final AddResult addResult = manifest.add( name,
                                                                          file );
                                reportAddResult( addResult,
                                                 logger,
                                                 project,
                                                 name );
                              }
                          }
                      }
                    else
                      {
                        logger.log( Level.INFO,
                                    "Skipping source inclusion due to top-level project setting: " + sourceHandling );
                      }
                    break;

                  }
              }
          }
        else
          {
            logger.log( Level.FINER,
                        "No sources" );
          }

        final List<Addition> additions = project.getAdditions();
        if( additions != null )
          {
            for( final Addition addition : additions )
              {
                final Where where = addition.where();
                if( (where == Where.BOTH) || (where == Where.JAR) )
                  {
                    try
                      {
                        manifest.add( addition.storeAs(),
                                      addition.getActualFile() );
                      }
                    catch( final IOException x )
                      {
                        logger.log( Level.SEVERE,
                                    "Cannot add '" + addition.getActualFile() + "' as '" + addition.storeAs() + "'",
                                    x );
                      }
                  }
              }
          }

        done.add( project );
        final List<ProjectReference> projectReferences = project.getProjects();
        if( projectReferences != null )
          {
            for( final ProjectReference projectRef : projectReferences )
              {
                final Project proj = projectRef.project();
                if( !done.contains( proj ) )
                  {
                    ArchiveHandlingOption handling = projectReferenceHandling.get( projectRef );
                    if( handling == null )
                      {
                        handling = projectRef.getHandling();
                      }
                    if( handling == ArchiveHandlingOption.IGNORE )
                      {
                        logger.log( Level.WARNING,
                                    "Ignoring project '" + proj.name() + "' archive: " + proj.getOutputJarFile() );
                        done.add( proj );
                      }
                    else if( handling == ArchiveHandlingOption.REFERENCE )
                      {
                        final String projPath = projectRef.getReferencePath().trim();
                        final String projOutput = new File( proj.getOutputJarFile() ).getName();
                        final String cpName;
                        if( projPath.length() == 0 )
                          {
                            cpName = projOutput;
                          }
                        else
                          {
                            cpName = new File( projPath,
                                               projOutput ).toString();
                          }
                        logger.log( Level.CONFIG,
                                    "Referencing project '" + proj.name() + "' archive: " + cpName );
                        manifest.add( cpName );
                        done.add( proj );
                      }
                    // NOTE: We defer MERGE until we have handled all projects,
                    // as sub-projects could define handling of the project to
                    // be MERGEd and mess up how we want to handle it.
                  }
              }

            for( final ProjectReference projectRef : projectReferences )
              {
                final Project proj = projectRef.project();
                if( !done.contains( proj ) )
                  {
                    ArchiveHandlingOption handling = projectReferenceHandling.get( projectRef );
                    if( handling == null )
                      {
                        handling = projectRef.getHandling();
                      }
                    if( handling == ArchiveHandlingOption.MERGE )
                      {
                        logger.log( Level.CONFIG,
                                    "Merging project '" + proj.name() + "' binaries" );
                        // prevent sub-project from referencing or merging again
                        done.add( proj );

                        // 'sourceHandling' is initially null (for the top-level
                        // project) so we used the project's own settings for
                        // each of potentially multiple source folders. For the
                        // sub-projects we want to include sources only if ALL
                        // top level sources are being included. This heuristic
                        // may be better addressed with an explicit flag later,
                        // but for now it makes enough sense to be workable.
                        FileHandlingOption srcHandling;
                        if( sourceHandling != null )
                          {
                            srcHandling = sourceHandling;
                          }
                        else
                          {
                            // Unless ALL our top-level sources are included,
                            // ignore subproject sources.
                            srcHandling = FileHandlingOption.INCLUDE;
                            for( final Sources sources : project.getSources() )
                              {
                                if( sources.getHandling() == FileHandlingOption.IGNORE )
                                  {
                                    srcHandling = FileHandlingOption.IGNORE;
                                    break;
                                  }
                              }
                          }

                        assemble( manifest,
                                  workingPath,
                                  proj,
                                  exclusions,
                                  sourceRoot,
                                  srcHandling,
                                  projectReferenceHandling,
                                  externalLibs,
                                  warnBinaryExclusion,
                                  done );
                      }
                  }
              }
          }
      }
    finally
      {
        logger.log( Level.INFO,
                    "Completed project " + project );
      }
  }


  private static List<Exclusion> allExclusions( final Project project )
  {
    final List<Exclusion> result = new ArrayList<>();

    final List<Exclusion> binaryExclusions = project.getBinariesExclusions();
    if( binaryExclusions != null )
      {
        result.addAll( binaryExclusions );
      }

    final List<Classpath> classpathList = project.getClasspath();
    if( classpathList != null )
      {
        for( final Classpath classpath : classpathList )
          {
            final List<Exclusion> classpathExclusions = classpath.getExclusions();
            if( classpathExclusions != null )
              {
                result.addAll( classpathExclusions );
              }
          }
      }

    final List<Sources> sourcesList = project.getSources();
    if( sourcesList != null )
      {
        for( final Sources sources : sourcesList )
          {
            if( sources.getHandling() == FileHandlingOption.INCLUDE )
              {
                final List<Exclusion> sourcesExclusions = sources.getExclusions();
                if( sourcesExclusions != null )
                  {
                    result.addAll( sourcesExclusions );
                  }
              }
          }
      }

    return result;
  }


  private static void reportAddResult( final AddResult addResult,
                                       final Logger logger,
                                       final Project project,
                                       final String name )
  {
    switch( addResult )
      {
      case CLONE:
        logger.log( Level.WARNING,
                    "Already exists under another name: Project " + project + " entry " + name );
        break;
      case CONFLICT:
        logger.log( Level.SEVERE,
                    "Conflicts with another file: Project " + project + " entry " + name );
        break;
      case DUPLICATE:
        logger.log( Level.INFO,
                    "Already added (from elsewhere): Project " + project + " entry " + name );
        break;
      case UNIQUE:
        logger.log( Level.FINE,
                    "Added: Project " + project + " entry " + name );
        break;
      }
  }


  /**
   * Produces a wrapped version of the line for the JAR's manifest so that the
   * line does not exceed 70 chars, and the continuation is indented by a space.
   *
   * @param line
   *          The line to be adjusted (if necessary)
   * @return The line that can be written to the manifest (the line may contain
   *         one or more newline characters).
   */
  private static String toJarLine( final String line )
  {
    final StringBuilder sb = new StringBuilder();
    final int len = line.length();
    int chunkLen = 70;
    int start = 0;
    do
      {
        if( start > 0 )
          {
            sb.append( " " );
            chunkLen = 69;
          }
        if( (start + chunkLen) < len )
          {
            sb.append( line.substring( start,
                                       start + chunkLen ) ).append( "\r\n" );
          }
        else
          {
            sb.append( line.substring( start ) ).append( "\r\n" );
          }
        start += chunkLen;
      }
    while( start < len );
    return sb.toString();
  }


  /**
   *
   * @param name
   * @param exclusions
   * @return 'true' if the file is to be omitted, 'false' if it is to be
   *         included.
   */
  private static boolean isExcluded( final String name,
                                     final List<Exclusion> exclusions )
  {
    if( exclusions != null )
      {
        for( final Exclusion exclusion : exclusions )
          {
            if( exclusion.isExcluded( name ) )
              {
                return true;
              }
          }
      }
    return false;
  }


  /**
   * Runs 'jarsigner' to determine whether we can sign JAR files with an
   * external command. If jarsigner is unavailable then we cannot sign the JAR
   * files we build, unless we were to re-implement jarsigner's relevant
   * functionality, which is currently out of scope.
   */
  public static void verifyJarSignerBinaryIsAvailable()
  {
    new Thread()
    {
      @Override
      public void run()
      {
        final String binary = Util.getSettings().get( "jarsigner-binary",
                                                      "jarsigner" );
        Util.logger().log( Level.FINEST,
                           "Looking for JarSigner binary" );
        try
          {
            // Don't care about return value, the 'jarSignerBinary' was either
            // set or it wasn't: We'll find out which when we later inquire
            // whether it worked.
            checkJarsigner( binary );
          }
        catch( final IOException x )
          {
            Util.logger().log( Level.SEVERE,
                               "JarSigner is not available",
                               x );
          }
      }
    }.start();
  }


  public static boolean checkJarsigner( final String binary )
    throws IOException
  {
    synchronized( jarSignerLock )
      {
        final OutputStream stdout = new ByteArrayOutputStream();

        final Runner runner = new Runner();
        runner.execute( null,
                        stdout,
                        null,
                        binary );

        final String out = stdout.toString();

        if( (out.indexOf( "-keystore" ) > -1) && (out.indexOf( "-storepass" ) > -1) && (out.indexOf( "-keypass" ) > -1) )
          {
            // We got 'usage' output from jarsigner, not an operating
            // system's error message, so we can invoke jarsigner just
            // fine
            jarSignerBinary = "jarsigner"; // it's available
            Util.logger().log( Level.FINEST,
                               "JarSigner found (signing JARs is supported)" );
            return true;
          }
        Util.logger().log( Level.SEVERE,
                           "JarSigner is not available" );
        return false;
      }
  }


  public static String getJarSignerBinary()
  {
    synchronized( jarSignerLock )
      {
        return jarSignerBinary;
      }
  }


  private Builder()
  {
  }

  private static String jarSignerBinary;
  private static Object jarSignerLock = new Object();
}
