package com.ringlord.boxer2;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.ringlord.Filename;
import com.ringlord.Settings;


/**
 * Attempt to execute parameters for the command line interface.
 *
 * @author K Udo Schuermann
 */
public class CLI
{
  public static void main( final String[] args )
  {
    go( new String[]{} );
  }


  /**
   * Process command line arguments.
   *
   * @param args
   *          Zero or more Command line parameters (must not be null).
   * @return A non-null and non-empty Map indicates that the GUI should be
   *         launched despite the presence of command line arguments. The
   *         following mappings are defined at this time:
   *         <ul>
   *         <li>"forceGUI" =&gt; String — A value "7" is to force a Swing GUI;
   *         a value "8" is to force a JavaFX&nbsp;8 GUI; other values presently
   *         should attempt to launch the most modern GUI first and fall back on
   *         progressively older ones until either successful or it is
   *         determined that no GUI is available (perhaps because of a headless
   *         configuration).
   *         <li>"projects" =&gt; {@link List}&lt;{@link Project}&gt; — A list
   *         of one of more Project objects resolved from their path names,
   *         which are to be built through the GUI rather than through the
   *         command line interface.
   *         </ul>
   */
  public static Map<String,Object> go( final String[] args )
  {
    final Map<String,Object> guiParams = new HashMap<>();
    final List<String> projectNames = new ArrayList<>();
    boolean verbose = false;
    boolean allowFlags = true;
    for( int i = 0; i < args.length; i++ )
      {
        final String arg = args[i];
        if( allowFlags && arg.startsWith( "-" ) )
          {
            if( arg.equals( "--" ) )
              {
                allowFlags = false;
              }
            // ==================================================
            else if( arg.equals( "-W" ) || arg.equals( "--workspace" ) )
              {
                if( ++i < args.length )
                  {
                    recordWorkspace( args[i] );
                  }
                else
                  {
                    showUsage();
                    System.err.println( "ERROR: Expected workspace directory following " + arg );
                    return null;
                  }
              }
            else if( arg.startsWith( "--workspace=" ) )
              {
                recordWorkspace( arg.substring( 12 ) );
              }
            // ==================================================
            else if( arg.equals( "-F" ) || arg.equals( "--forget" ) )
              {
                if( ++i < args.length )
                  {
                    forgetWorkspace( args[i] );
                  }
                else
                  {
                    showUsage();
                    System.err.println( "ERROR: Expected workspace directory following " + arg );
                    return null;
                  }
              }
            else if( arg.startsWith( "--forget=" ) )
              {
                forgetWorkspace( arg.substring( 9 ) );
              }
            // ==================================================
            else if( arg.equals( "-G" ) || arg.equals( "--gui" ) )
              {
                if( guiParams.get( "forceGUI" ) == null )
                  {
                    guiParams.put( "forceGUI",
                                   "" );
                  }
              }
            else if( arg.equals( "-7" ) )
              {
                guiParams.put( "forceGUI",
                               "7" );
              }
            else if( arg.equals( "-8" ) )
              {
                guiParams.put( "forceGUI",
                               "8" );
              }
            // ==================================================
            else if( arg.equals( "-v" ) || arg.equals( "--verbose" ) )
              {
                verbose = true;
              }
            // ==================================================
            else if( arg.equals( "-X" ) || arg.equals( "--extract" ) )
              {
                showCopyright();
                extractFile( "boxer.pdf" );
                extractFile( "COPYING" );
              }
            // ==================================================
            else if( arg.equals( "-H" ) || arg.equals( "--help" ) )
              {
                showUsage();
              }
            // ==================================================
            else if( arg.equals( "-L" ) || arg.equals( "--license" ) )
              {
                showLicense();
              }
            // ==================================================
            else if( arg.equals( "-C" ) || arg.equals( "--copyright" ) )
              {
                showCopyright();
              }
            // ==================================================
            else if( arg.equals( "-V" ) || arg.equals( "--version" ) )
              {
                showVersion();
              }
            // ==================================================
            else
              {
                showUsage();
                System.err.println( "ERROR: Unrecognized flag: " + arg );
              }
          }
        else
          {
            projectNames.add( arg );
          }
      }

    final List<Project> projects;
    if( projectNames.isEmpty() )
      {
        projects = null;
      }
    else
      {
        try
          {
            projects = Util.findProjects( projectNames );
          }
        catch( final IllegalArgumentException x )
          {
            showUsage();
            System.err.println( x.getMessage() );
            return null;
          }
      }

    if( !guiParams.isEmpty() )
      {
        if( projects != null )
          {
            guiParams.put( "projects",
                           projects );
          }
        final Settings settings = Util.getSettings();
        final String forceGuiVersion = settings.get( "forceGUI",
                                                     (String)null );
        if( forceGuiVersion != null )
          {
            guiParams.put( "forceGUI",
                           forceGuiVersion );
          }
        if( versionShown && !headerShown )
          {
            System.err.println();
          }
        return guiParams;
      }

    if( projects == null )
      {
        if( !versionShown )
          {
            // No part of the usage was shown, so show it all
            showUsage();
          }
      }
    else
      {
        final Logger logger = Util.logger();

        final ConsoleHandler consoleHandler = new ConsoleHandler();
        consoleHandler.setLevel( verbose
            ? Level.CONFIG
            : Level.INFO );
        consoleHandler.setFormatter( new Util.PlainLog() );
        logger.addHandler( consoleHandler );

        final long startTimeAll = System.currentTimeMillis();
        for( final Project project : projects )
          {
            final long startTime = System.currentTimeMillis();
            final long buildNumber = project.bumpBuildNumber();
            logger.log( Level.INFO,
                        "Processing " + project + " (Build " + buildNumber + ")" );
            if( project.isSubproject() )
              {
                logger.log( Level.SEVERE,
                            "Project is marked as a 'subproject' and is not buildable by itself." );
                continue;
              }

            final File archive = Util.resolveFilename( project,
                                                       project.getOutputJarFile() );
            final File distro = Util.resolveFilename( project,
                                                      project.getDistributionFile() );
            final File workingPath = Filename.resolveTilde( project.getWorkingDirectory() );

            try
              {
                // External library file references that may need to be package
                // into the distribution archive
                final List<Project.Library> externalLibs = new ArrayList<>();
                final Manifest manifest = Builder.assemble( project,
                                                            workingPath,
                                                            externalLibs );

                Builder.writeJAR( project,
                                  archive,
                                  manifest );
                Builder.signJAR( project,
                                 archive,
                                 null );
                try
                  {
                    archive.setExecutable( true );
                  }
                catch( final Throwable x )
                  {
                    x.printStackTrace();
                  }
                Builder.writeJNLP( project,
                                   archive );
                Builder.writeDistro( distro,
                                     archive,
                                     Util.resolveFilename( project,
                                                           project.getDistributionSubdir() ).toString(),
                                     externalLibs,
                                     project.getAdditions() );
                Builder.writeDigestFiles( project,
                                          archive,
                                          distro );
              }
            catch( NoSuchAlgorithmException |
                   IOException e )
              {
                // TODO Auto-generated catch block
                e.printStackTrace();
              }
            final long totalTime = System.currentTimeMillis() - startTime;
            logger.log( Level.INFO,
                        String.format( "Completed project " + project + " in %1.3f seconds",
                                       totalTime / 1000.0d ) );
            project.save();
          }

        if( projects.size() > 1 )
          {
            final long totalTimeAll = System.currentTimeMillis() - startTimeAll;
            logger.log( Level.INFO,
                        String.format( "Completed all projects in %1.3f seconds",
                                       totalTimeAll / 1000.0d ) );
          }
      }

    if( versionShown && !headerShown )
      {
        System.err.println();
      }
    return null;
  }


  private static void extractFile( final String name )
  {
    final File outFile = new File( name ).getAbsoluteFile();
    try( final InputStream iStream = CLI.class.getClassLoader().getResourceAsStream( name ) )
      {
        if( iStream == null )
          {
            Util.logger().log( Level.WARNING,
                               "No such file to extract: " + name );
            System.err.println( "No such file to extract: " + name );
          }
        else
          {
            try( final OutputStream oStream = new FileOutputStream( outFile ) )
              {
                Util.copy( iStream,
                           oStream );
                Util.logger().log( Level.INFO,
                                   "Extracted file: " + outFile );
                System.err.println( "Extracted file: " + outFile );
              }
          }
      }
    catch( final IOException x )
      {
        Util.logger().log( Level.SEVERE,
                           "Cannot extract file: " + outFile );
        System.err.println( "Cannot extract file: " + outFile );
      }
  }


  private static void recordWorkspace( final String path )
  {
    final File name = new File( path );
    if( name.isDirectory() )
      {
        showCopyright();
        System.err.println( "Would now record '" + name + "' as a known workspace" );
      }
  }


  private static void forgetWorkspace( final String path )
  {
    final File name = new File( path );
    if( name.isDirectory() )
      {
        showCopyright();
        System.err.println( "Would now forget about the workspace '" + name + "'" );
      }
  }


  private static void showVersion()
  {
    if( !versionShown )
      {
        versionShown = true;
        System.err.println( "Boxer " + Util.VERSION + " (" + Util.VERDATE + ")" );
      }
  }


  private static void showCopyright()
  {
    if( !headerShown )
      {
        headerShown = true;

        showVersion();
        System.err.println( "Copyright \u00a9 " + Util.CPYEARS + " by K. Udo Schuermann" );
        System.err.println( "All rights reserved" );
        System.err.println();
      }
  }


  private static void showLicense()
  {
    if( !licenseShown )
      {
        licenseShown = true;

        showCopyright();
        System.err.println( "This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you" );
        System.err.println( "are welcome to redistribute it under certain conditions. Please refer to the" );
        System.err.println( "GNU General Public License v3 (the file COPYING in the .jar file) for details." );
        System.err.println();
      }
  }


  private static void showUsage()
  {
    if( !usageShown )
      {
        usageShown = true;

        showLicense();
        System.err.println( "Usage: java -jar boxer.jar [ parameters ] [ project [ ...]]" );
        System.err.println( "where parameters are:" );
        System.err.println( "  -W path           Register the named directory as a workspace within which" );
        System.err.println( "  --workspace path  project directories are to be found. See also -F/--forget" );
        System.err.println();
        System.err.println( "  -F path           Forget the named workspace directory. This has no effect if" );
        System.err.println( "  --forget path     this directory was not registered using -W/--workspace or" );
        System.err.println( "                    the equivalent function from the GUI." );
        System.err.println();
        System.err.println( "  -G                Start the GUI even though command line parameters (such as" );
        System.err.println( "  --gui             project name(s), see below) are present. Attempts to launch" );
        System.err.println( "                    the most modern GUI first, and falls back on progressively" );
        System.err.println( "                    older styles. See also -7 and -8 parameters." );
        System.err.println();
        System.err.println( "  -7                Force the use of Java 7 (or earlier) Swing-style GUI even if" );
        System.err.println( "                    JavaFX 8 is available. This implies (makes unnecessary) the" );
        System.err.println( "                    -G/--gui parameter." );
        System.err.println();
        System.err.println( "  -8                Force the use of Java 8 (or later) JavaFX-style GUI, though" );
        System.err.println( "                    it is the default if you have Java 8+ installed and cannot" );
        System.err.println( "                    work if you don't. This implies (makes unnecessary) the -G/" );
        System.err.println( "                    --gui parameter. COMING IN A FUTURE VERSION OF BOXER 2.x" );
        System.err.println();
        System.err.println( "  -v                Be more verbose about the processing, which can be of great" );
        System.err.println( "  --verbose         help if you are trying to diagnose issues. The log file" );
        System.err.println( "                    named " + new File( System.getProperty( "java.io.tmpdir" ),
                                                                     "boxer-00.log" ) );
        System.err.println( "                    captures the most information of all, including timings," );
        System.err.println( "                    and will always be verbose." );
        System.err.println();
        System.err.println( "  -X                Extract the documentation file (boxer.pdf) and GNU General" );
        System.err.println( "  --extract         Public License v3 (COPYING) to the current working" );
        System.err.println( "                    directory. Note that the documentation is critical if you" );
        System.err.println( "                    wish to build or modify the project-specific .boxer2-conf" );
        System.err.println( "                    files manually, rather than through the GUI (which does it" );
        System.err.println( "                    all for you)." );
        System.err.println();
        System.err.println( "  -V / --version    Display the version and exit." );
        System.err.println();
        System.err.println( "  -C / --copyright  Display the copyright header and exit." );
        System.err.println();
        System.err.println( "  -L / --license    Display copyright header and license, then exit." );
        System.err.println();
        System.err.println( "  -H / --help       Display this message and exit." );
        System.err.println();
        System.err.println( "and project are one or more project names or uniquely identifiable portions of" );
        System.err.println( "project names to be built using the configuration in the .boxer2-conf file in" );
        System.err.println( "each named project's directory. The .boxer2-conf file is edited through the GUI," );
        System.err.println( "but can be modified (with care) by hand: For details on the format of this file" );
        System.err.println( "please see the documentation, which can be generated/extracted to a file using" );
        System.err.println( "the -X/--extract parameter or downloaded from http://ringlord.com/boxer.html" );
        System.err.println( "     Note that an exact project name match will take precedence over a partial" );
        System.err.println( "match, so with three projects, 'test', 'test1a', and 'test1b' giving the name" );
        System.err.println( "'test' will cleanly match the first one, but 'test1' will be ambiguous." );
        System.err.println();
      }
  }
  private static boolean versionShown;
  private static boolean headerShown;
  private static boolean licenseShown;
  private static boolean usageShown;
}
