package com.ringlord.boxer2;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.ringlord.boxer2.Project.Library;


/**
 * A collection of {@link File} objects whose content is to be recorded under a
 * given name in an archive or similar collection.
 *
 * @author K Udo Schuermann
 */
public class Manifest
  implements
    Iterable<String>
{
  public Manifest()
  {
    super();
    this.filenameComparator = new FilenameComparator( DEFAULT_EXTENSIONS );
  }


  public void setMainClassName( final String mainClassName )
  {
    this.mainClassName = mainClassName;
  }


  public String getMainClassName()
  {
    return mainClassName;
  }


  @Override
  public Iterator<String> iterator()
  {
    final List<String> names = new ArrayList<>();
    names.addAll( content.keySet() );
    Collections.sort( names,
                      filenameComparator );
    return names.iterator();
  }


  /**
   * Obtains the {@link File} that was added under a given name.
   *
   * @param nameInPackage
   *          The name of the file, it is case sensitive. The name will be
   *          normalized internally, converting all backslashes ('\') are to
   *          forward-leaning directory slashes, and ensuring that the name
   *          begins with a '/'
   * @return The File associated with the given name
   * @throws NullPointerException
   *           No File was {@link #add(File, String) added} under that name.
   */
  public File getFile( final String nameInPackage )
  {
    final String name = normalizeName( nameInPackage );
    return content.get( name ).file;
  }


  public List<String> getClassPathEntries()
  {
    final List<String> result = new ArrayList<>();
    result.addAll( classpath );
    return result;
  }


  public int fileCount()
  {
    return fileCount;
  }


  public long byteCount()
  {
    return byteCount;
  }


  public void add( final String entry )
  {
    classpath.add( entry );
  }


  public AddResult add( final File workingDirectory,
                        final String nameInPackage,
                        final Library library )
    throws IOException
  {
    try
      {
        final File workingFile = new File( workingDirectory,
                                           nameInPackage );
        workingFile.getParentFile().mkdirs();
        try( final OutputStream f = new FileOutputStream( workingFile ) )
          {
            if( library.copyTo( nameInPackage,
                                f ) )
              {
                fileCount++;
                byteCount += workingFile.length();
                return add( nameInPackage,
                            workingFile );
              }
          }
        throw new IOException( "Could not copy " +
                               nameInPackage +
                               " from " +
                               library.toString() +
                               " to " +
                               workingDirectory );
      }
    catch( final Throwable x )
      {
        final IOException boom = new IOException( "Could not copy " +
                                                  nameInPackage +
                                                  " from " +
                                                  library.toString() +
                                                  " to " +
                                                  workingDirectory );
        boom.initCause( x );
        throw boom;
      }
  }


  /**
   * @param file
   *          The File to be added
   * @param nameInPackage
   *          The name of the file, it is case sensitive. The name will be
   *          normalized internally, converting all backslashes ('\') are to
   *          forward-leaning directory slashes, and ensuring that the name
   *          begins with a '/'
   * @return One of four possible {@link AddResult results}, of which only one,
   *         {@link AddResult.CONFLICT}, represents an actual error condition.
   */
  public AddResult add( final String nameInPackage,
                        final File file )
    throws NoSuchAlgorithmException,
      IOException
  {
    final String name = normalizeName( nameInPackage );
    final String signature1 = getMD1( file );
    String signature2 = null;

    String existingName = null;

    for( final Map.Entry<String,FileReference> e : content.entrySet() )
      {
        final FileReference f = e.getValue();

        // equality test is faster than signature computation
        if( f.file.equals( file ) )
          {
            existingName = e.getKey();
            break;
          }

        if( signature1.equals( f.signature1() ) )
          {
            if( signature2 == null )
              {
                signature2 = getMD2( file );
              }
            if( signature2.equals( f.signature2() ) )
              {
                existingName = e.getKey();
                break;
              }
          }
      }

    if( existingName != null )
      {
        if( existingName.equals( name ) )
          {
            // We are adding the exact same content (possibly identical file)
            // under the same name, so it is totally safe to do nothing at all
            return AddResult.DUPLICATE;
          }
        // We have the same content but we are adding it under a different name;
        // this is rather wasteful but otherwise not harmful.
        content.put( name,
                     new FileReference( file,
                                        signature1,
                                        signature2 ) );
        fileCount++;
        byteCount += file.length();
        return AddResult.CLONE;
      }
    else if( content.get( name ) != null )
      {
        // We have unique content under a name that exists already. This is
        // really bad, as there is no way to determine (automatically) which
        // file is to be kept and which to be lost/discarded.
        return AddResult.CONFLICT;
      }

    // We have a unique file under a unique name
    content.put( name,
                 new FileReference( file,
                                    signature1,
                                    signature2 ) );
    fileCount++;
    byteCount += file.length();
    return AddResult.UNIQUE;
  }


  private static String normalizeName( final String name )
  {
    String s = name.replace( '\\',
                             '/' );
    if( s.startsWith( "/" ) )
      {
        s = s.substring( 1 );
      }
    return s;
  }


  private static String getMD1( final File file )
    throws NoSuchAlgorithmException,
      IOException
  {
    return getMD( file,
                  1024,
                  MessageDigest.getInstance( "MD5" ) );
  }


  private static String getMD2( final File file )
    throws NoSuchAlgorithmException,
      IOException
  {
    return getMD( file,
                  Long.MAX_VALUE,
                  MessageDigest.getInstance( "SHA-256" ) );
  }


  private static String getMD( final File file,
                               final long size,
                               final MessageDigest md )
    throws IOException
  {
    try( final BufferedInputStream f = new BufferedInputStream( new FileInputStream( file ) ) )
      {

        final byte[] buffer = new byte[4096];
        int inBuffer = 0;
        long remaining = Math.min( size,
                                   file.length() );
        while( (remaining > 0) && ((inBuffer = f.read( buffer,
                                                       0,
                                                       Math.min( (int)remaining,
                                                                 buffer.length ) )) > -1) )
          {
            md.update( buffer,
                       0,
                       inBuffer );
            remaining -= inBuffer;
          }
      }
    return byteArrayToString( md.digest() );
  }


  private static final String byteArrayToString( final byte[] data )
  {
    final StringBuilder sb = new StringBuilder();
    for( int i = 0; i < data.length; i++ )
      {
        final int n = data[i];
        sb.append( nibble[(n >> 4) & 0x0f] ).append( nibble[n & 0x0f] );
      }
    return sb.toString();
  }


  public enum AddResult
  {
    /**
     * Unique content under a unique name (this is the "best" situation)
     */
    UNIQUE,
    /**
     * Unique content whose name conflicts with an existing name (this is bad,
     * because two unique files cannot have the same name)
     */
    CONFLICT,
    /**
     * Identical content under a different name (wasteful but likely needed)
     */
    DUPLICATE,
    /**
     * Identical content under identical name (skipped)
     */
    CLONE;
  }


  @Override
  public String toString()
  {
    final StringBuilder sb = new StringBuilder();
    for( final String s : classpath )
      {
        sb.append( "\tClasspath: " + s + "\n" );
      }
    for( final Map.Entry<String,FileReference> e : content.entrySet() )
      {
        sb.append( "\tStore: " )
          .append( e.getValue().file.toString() )
          .append( "\n" )
          .append( "\t   as: " )
          .append( e.getKey() )
          .append( "\n" );
      }
    return sb.toString();
  }


  private class FileReference
  {
    FileReference( final File file,
                   final String signature1,
                   final String signature2 )
    {
      super();
      this.file = file;
      this.signature1 = signature1;
      this.signature2 = signature2;
    }


    String signature1()
    {
      return signature1;
    }


    String signature2()
      throws NoSuchAlgorithmException,
        IOException
    {
      if( signature2 == null )
        {
          signature2 = getMD2( file );
        }
      return signature2;
    }

    private String signature2;
    private final String signature1;
    private final File file;
  }


  class FilenameComparator
    implements
      Comparator<String>
  {
    FilenameComparator( final List<String> extensions )
    {
      super();
      this.extensions = (extensions == null
          ? new ArrayList<String>()
          : extensions);
    }


    @Override
    public int compare( final String name1,
                        final String name2 )
    {
      final int lastDot1 = name1.lastIndexOf( "." );
      final int lastDot2 = name2.lastIndexOf( "." );

      final String ext1 = (lastDot1 < 0
          ? "" // sort to the end
          : name1.substring( lastDot1 )).toLowerCase();
      final String ext2 = (lastDot2 < 0
          ? "" // sort to the end
          : name2.substring( lastDot2 )).toLowerCase();

      if( !ext1.equals( ext2 ) )
        {
          // The extensions are not the same, so try to use the specific
          // ordering we've picked
          final int index1 = extensions.indexOf( ext1 );
          final int index2 = extensions.indexOf( ext2 );
          if( (index1 > -1) && (index2 > -1) )
            {
              return index1 - index2;
            }
        }

      // Extensions are the same or we cannot compare the positions reasonably
      return name1.compareTo( name2 );
    }
    final private List<String> extensions;
  }

  private String mainClassName;
  private int fileCount;
  private long byteCount;
  //
  private final List<String> classpath = new ArrayList<>();
  private final Map<String,FileReference> content = new HashMap<>();
  private final FilenameComparator filenameComparator;
  /**
   * Our preferred ordering of files in the output JAR.
   */
  private static final List<String> DEFAULT_EXTENSIONS = new ArrayList<String>()
  {
      {
        add( ".class" ); // classes have highest priority
        add( ".png" );
        add( ".jpeg" );
        add( ".jpg" );
        add( ".gif" );
        add( ".tiff" );
        add( ".bmp" );
        add( ".dat" );
        add( ".data" );
        add( ".bin" );
        add( ".html" );
        add( ".java" );
        add( ".text" );
        add( ".txt" );
        add( ".tex" );
        add( ".dvi" );
        add( ".ps" );
        add( ".doc" );
        add( ".docx" );
        add( ".pdf" );
        add( ".odt" );
        add( "" );
      }
    private static final long serialVersionUID = 6125253628595838640L;
  };
  private static final char[] nibble = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
}
