package com.ringlord.boxer2;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.imageio.ImageIO;
import javax.swing.Icon;
import javax.swing.ImageIcon;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.ringlord.Config;
import com.ringlord.boxer2.Project.Addition.Where;
import com.ringlord.mime.Base64;


/**
 * Represents an Eclipse project within a given workspace. The project must
 * contain a <tt>.classpath</tt> file at the top level.
 *
 * @author K Udo Schuermann
 */
public class Project
  implements
    Comparable<Project>,
    Closeable
{
  /**
   * Construct a Project object based on an existing named project in the given
   * {@link Workspace}. Boxer expects the top level project directory to contain
   * an XML file named <tt>.classpath</tt>, which describes the project's
   * components, such as output path for compiled binaries (classes), libraries,
   * source files, and project references.
   *
   * @param workspace
   *          The {@link Workspace} for the project.
   * @param name
   *          The name of the project. This is the directory name within the
   *          Workspace, and is case sensitive.
   * @throws SAXException
   *           The <tt>.classpath</tt> file exists, but could not be parsed.
   * @throws IOException
   *           The <tt>.classpath</tt> file exists but could not be opened.
   * @throws IllegalArgumentException
   *           The <tt>name</tt> does not actually refer to a project (no
   *           <tt>.classpath</tt> file found).
   */
  Project( final Workspace workspace,
           final String name )
    throws SAXException,
      IOException
  {
    super();
    this.workspace = workspace;
    this.projectName = name;
    this.projectRoot = new File( workspace.root(),
                                 name );

    this.config = new Config( new File( projectRoot,
                                        ".boxer2-conf" ),
                              "For information about Boxer, visit http://ringlord.com/boxer.html" );

    final File classpathFile = new File( projectRoot,
                                         ".classpath" );
    if( classpathFile.isFile() )
      {
        workspace.saxParser().parse( classpathFile,
                                     new ClasspathXmlHandler() );
      }
    else
      {
        throw new IllegalArgumentException( "Not a project (" +
                                            name +
                                            ") in workspace (" +
                                            workspace.root() +
                                            "): No file " +
                                            classpathFile );
      }
    this.checkedEntryPoints = false;
  }


  public void save()
  {
    config.save();
  }


  @Override
  public void close()
  {
    save();
  }


  public void reset()
  {
    config.reset();
  }


  /**
   * The name of the project, equivalent to the name of the directory in which
   * it resides. The name is case sensitive.
   *
   * @return The name of the project.
   */
  public String name()
  {
    return projectName;
  }


  /**
   * The directory where the project resides.
   *
   * @return The directory of the project.
   */
  public File root()
  {
    return projectRoot;
  }


  public void setIcon( final ImageIcon icon )
  {
    if( icon == null )
      {
        config.put( "project-icon-base64",
                    (String)null );
        this.projectIcon = null;
      }
    else
      {
        final Image image = icon.getImage().getScaledInstance( 16,
                                                               16,
                                                               Image.SCALE_SMOOTH );
        final BufferedImage bufferedImage = new BufferedImage( 16,
                                                               16,
                                                               BufferedImage.TYPE_INT_RGB );
        bufferedImage.getGraphics().drawImage( image,
                                               0,
                                               0,
                                               null );
        final ByteArrayOutputStream iconData = new ByteArrayOutputStream();
        try
          {
            ImageIO.write( bufferedImage,
                           "png",
                           iconData );
            final byte[] bytesPNG = iconData.toByteArray();
            final byte[] base64 = Base64.encode( bytesPNG );
            config.put( "project-icon-base64",
                        new String( base64 ) );
            this.projectIcon = new ImageIcon( bufferedImage );
          }
        catch( final IOException x )
          {
            Util.logger().log( Level.SEVERE,
                               "Cannot process project icon",
                               x );
          }
      }

    mustLoadProjectIcon = false;
  }


  public Icon getIcon()
  {
    if( mustLoadProjectIcon )
      {
        mustLoadProjectIcon = false;
        final String s = config.get( "project-icon-base64",
                                     (String)null );
        if( s != null )
          {
            final byte[] base64 = s.getBytes();
            final byte[] bytesPNG = Base64.decode( base64 );
            this.projectIcon = new ImageIcon( bytesPNG );
          }
      }
    return projectIcon;
  }


  public void setDescription( final String description )
  {
    config.put( "description",
                description );
  }


  public String getDescription()
  {
    return config.get( "description",
                       "" );
  }


  public void setSubproject( final boolean isSubproject )
  {
    config.put( "subproject-not-independent",
                isSubproject );
  }


  public boolean isSubproject()
  {
    return config.get( "subproject-not-independent",
                       false );
  }


  public void setSplashImageName( final String splashImageName )
  {
    config.put( "splash-image",
                splashImageName );
  }


  public String getSplashImageName()
  {
    return config.get( "splash-image",
                       (String)null );
  }


  public void setWorkingDirectory( final String workingDirectory )
  {
    config.put( "outpath",
                workingDirectory );
  }


  public String getWorkingDirectory()
  {
    return config.get( "outpath",
                       System.getProperty( "java.io.tmpdir" ) );
  }


  public void setOutputJarName( final String outputJarName )
  {
    config.put( "jar-file",
                outputJarName );
  }


  public String getOutputJarFile()
  {
    final String outputJarName = config.get( "jar-file",
                                             (String)null );
    return (outputJarName == null
        ? new File( root(),
                    projectName + ".jar" ).toString()
        : outputJarName);
  }


  public void setCreateDistributionFile( final boolean createDistributionFile )
  {
    config.put( "create-distro-file",
                createDistributionFile );
  }


  public boolean createDistributionFile()
  {
    return config.get( "create-distro-file",
                       false );
  }


  public void setDistributionName( final String distributionFile )
  {
    config.put( "distro-file",
                distributionFile );
  }


  public String getDistributionFile()
  {
    final String distributionName = config.get( "distro-file",
                                                (String)null );
    return (distributionName == null
        ? new File( getWorkingDirectory(),
                    projectName + ".tar.gz" ).toString()
        : distributionName);
  }


  public void setDistributionSubdir( final String subdir )
  {
    config.put( "distro-subdir",
                subdir );
  }


  public String getDistributionSubdir()
  {
    return config.get( "distro-subdir",
                       projectName );
  }


  public List<String> getDigestsPicked()
  {
    final List<String> result = config.get( "create-digest-files",
                                            (List<String>)null,
                                            ";" );
    return (result == null
        ? new ArrayList<String>()
        : result);
  }


  public void setDigestsPicked( final List<String> pickedDigests )
  {
    config.put( "create-digest-files",
                pickedDigests,
                ";" );
  }


  public boolean isDigestPicked( final String digestName )
  {
    final List<String> pickedDigests = getDigestsPicked();
    if( pickedDigests != null )
      {
        return pickedDigests.contains( digestName );
      }
    return false;
  }


  public void setDigestPicked( final String digestName,
                               final boolean isDigestPicked )
  {
    if( (digestName == null) || (digestName.length() == 0) )
      {
        return;
      }
    List<String> pickedDigests = getDigestsPicked();
    if( pickedDigests == null )
      {
        pickedDigests = new ArrayList<>();
      }
    if( isDigestPicked )
      {
        if( !pickedDigests.contains( digestName ) )
          {
            pickedDigests.add( digestName );
          }
      }
    else
      {
        pickedDigests.remove( digestName );
      }
    setDigestsPicked( pickedDigests );
  }


  public void setEntryPointName( final String entryPointName )
  {
    config.put( "main-class",
                entryPointName );
  }


  public String getEntryPointName()
  {
    String name = config.get( "main-class",
                              (String)null );
    if( (name == null) && !checkedEntryPoints )
      {
        // This method will update the config but it also returns the list, so
        // it's faster to grab that value directly than check the config again
        final List<Class<?>> entryPoints = getEntryPoints();
        if( entryPoints.size() == 1 )
          {
            name = entryPoints.get( 0 ).getName();
          }
      }
    return name;
  }


  // ==================================================
  // Begin JNLP-related

  public void setBuildJNLP( final boolean buildJNLP )
  {
    config.put( "build-jnlp",
                buildJNLP );
  }


  public boolean getBuildJNLP()
  {
    return config.get( "build-jnlp",
                       false );
  }


  public void setJNLP( final String jnlpXML )
  {
    final byte[] data = Base64.encode( jnlpXML.getBytes( UTF8 ) );
    config.put( "jnlp-base64",
                new String( data ) );
  }


  public String getJNLP()
  {
    final String xml = config.get( "jnlp-base64",
                                   (String)null );
    if( xml != null )
      {
        try
          {
            final byte[] data = Base64.decode( xml.getBytes( UTF8 ) );
            return new String( data,
                               UTF8 );
          }
        catch( final Exception x )
          {
            Util.logger().log( Level.SEVERE,
                               "Failed to convert Base64-decoded data into UTF-8 String",
                               x );
          }
      }
    return defaultJNLP;
  }


  // End JNLP-related
  // ==================================================

  public void setSignJAR( final boolean signJar )
  {
    config.put( "sign-jar",
                signJar );
  }


  public boolean getSignJAR()
  {
    return config.get( "sign-jar",
                       false );
  }


  public void setSigningPropertiesFile( final String signingPropertiesFile )
  {
    config.put( "sign-properties-file",
                signingPropertiesFile );
  }


  public String getSigningPropertiesFile()
  {
    return config.get( "sign-properties-file",
                       Util.getSettings().get( "sign-properties-file-default",
                                               new File( "~/.keystore-info" ).toString() ) );
  }


  /**
   * Obtain classes with <code>public static void main(String[])</code> methods.
   * At this time it is not possible to include classes with unmet external
   * dependencies as a complete project classpath is not being constructed for
   * this operation.
   *
   * @return A list of zero or more classes that can serve as the program entry
   *         point.
   */
  public List<Class<?>> getEntryPoints()
  {
    final List<Class<?>> javaClasses = new ArrayList<>();
    for( final Binaries binaries : getBinaries() )
      {
        EntryPointCollector.collect( javaClasses,
                                     binaries.root() );
      }

    final List<Class<?>> results = new ArrayList<>();
    boolean explainUnmetDependency = false;
    for( final Class<?> c : javaClasses )
      {
        try
          {
            final Method m = c.getDeclaredMethod( "main",
                                                  String[].class );
            if( ((m.getModifiers() & PUBLIC_STATIC) == PUBLIC_STATIC) && m.getReturnType().equals( Void.TYPE ) )
              {
                Util.logger().log( Level.INFO,
                                   "Entry point found: " + c );
                results.add( c );
              }
          }
        catch( final NoSuchMethodException e )
          {
            // No public static void main( String[] )
          }
        catch( final NoClassDefFoundError e )
          {
            /**
             * TODO Reference ALL classes (Binaries, Library, Classpath, AND
             * recursively those of dependent project's, as well!) ... until
             * then we can and will suffer unresolved link errors.
             */
            Util.logger().log( Level.WARNING,
                               "Cannot inspect " + c + " (depends on " + e.getMessage() + ")" );
            explainUnmetDependency = true;
          }
        catch( final SecurityException e )
          {
            // The class is not public, so we cannot inspect its methods
          }
      }

    if( explainUnmetDependency )
      {
        Util.logger().log( Level.WARNING,
                           "The above listed issues about being unable to inspect "
                               + "classes in search of 'public static void main(String[])' "
                               + "entry points result from the fact that each class is "
                               + "inspected in isolation and not in the context of "
                               + "dependent classes, libraries, and referenced project "
                               + "elements available, too. Doing this is currently "
                               + "beyond Boxer's scope, but its implementation is under "
                               + "consideration." );
      }

    if( !checkedEntryPoints )
      {
        checkedEntryPoints = true;
        if( (results.size() == 1) && (getEntryPointName() == null) )
          {
            setEntryPointName( results.get( 0 ).getName() );
          }
      }
    return results;
  }


  public void setVersionExtractFilename( final String filename )
  {
    config.put( "version-extract-filename",
                filename );
  }


  public String getVersionExtractFilename()
  {
    // return non-null default to avoid null pointer exception on new File()
    return config.get( "version-extract-filename",
                       "" );
  }


  public void setVersionExtractRegexes( final List<String> regex )
  {
    int i = 0;
    for( final String s : regex )
      {
        config.put( "version-extract-regex." + i,
                    s );
        ++i;
      }
    // delete all others beyond what we are trying to save
    while( true )
      {
        final String key = "version-extract-regex." + i;
        if( config.get( key,
                        (String)null ) == null )
          {
            break;
          }
        config.put( key,
                    (String)null );
        ++i;
      }
  }


  public List<String> getVersionExtractRegexes()
  {
    final List<String> result = new ArrayList<>();
    int i = 1;
    while( true )
      {
        final String value = config.get( "version-extract-regex." + i,
                                         (String)null );
        if( value == null )
          {
            break;
          }
        result.add( value );
        ++i;
      }
    if( result.isEmpty() )
      {
        result.add( Util.getSettings().get( "version-sample-regex",
                                            "VERSION\\s*=\\s*\"(.*?)\"" ) );
      }
    return result;
  }


  public void setVersionExtractAssemblies( final List<String> spec )
  {
    int i = 0;
    for( final String s : spec )
      {
        config.put( "version-extract-assembly." + i,
                    s );
        ++i;
      }
    // delete all others beyond what we are trying to save
    while( true )
      {
        final String key = "version-extract-assembly." + i;
        if( config.get( key,
                        (String)null ) == null )
          {
            break;
          }
        config.put( key,
                    (String)null );
        ++i;
      }
  }


  public List<String> getVersionExtractAssemblies()
  {
    final List<String> result = new ArrayList<>();
    int i = 1;
    while( true )
      {
        final String value = config.get( "version-extract-assembly." + i,
                                         (String)null );
        if( value == null )
          {
            break;
          }
        result.add( value );
        ++i;
      }
    if( result.isEmpty() )
      {
        result.add( Util.getSettings().get( "version-sample-assembly",
                                            "$(1)" ) );
      }
    return result;
  }


  /**
   * Obtains the build number for this next build; this number will become
   * available as ${BUILD} and is immediately updated in the project config so
   * that the next call to this method gets the next value. If undefined or less
   * than 1, the value will default to 1.
   */
  public long bumpBuildNumber()
  {
    long buildNumber = config.get( "last-build-number",
                                   0L );
    if( buildNumber <= 0 )
      {
        buildNumber = 1;
      }
    else
      {
        buildNumber++;
      }
    config.put( "last-build-number",
                buildNumber );
    return buildNumber;
  }


  public long currentBuildNumber()
  {
    long buildNumber = config.get( "last-build-number",
                                   0L );
    if( buildNumber <= 0 )
      {
        buildNumber = bumpBuildNumber();
      }
    return buildNumber;
  }


  public void setBinariesExclusions( final List<Exclusion> exclusions )
  {
    if( this.binariesExclusions == null )
      {
        this.binariesExclusions = new ArrayList<>();
      }
    else
      {
        this.binariesExclusions.clear();
      }
    this.binariesExclusions.addAll( exclusions );
    saveExclusions( "binaries-exclusion",
                    exclusions );
  }


  public List<Exclusion> getBinariesExclusions()
  {
    if( this.binariesExclusions == null )
      {
        this.binariesExclusions = new ArrayList<>();
        loadExclusions( "binaries-exclusion",
                        this.binariesExclusions );
      }
    final List<Exclusion> result = new ArrayList<>();
    result.addAll( this.binariesExclusions );
    return result;
  }


  public void setSourceRoot( final String root )
  {
    config.put( "source-root",
                root );
  }


  public String getSourceRoot()
  {
    return config.get( "source-root",
                       "source" );
  }


  public void setManifestAdditions( final List<String> manifestAdditions )
  {
    config.put( "jar-manifest-addition",
                manifestAdditions );
  }


  public List<String> getManifestAdditions()
  {
    final List<String> result = config.get( "jar-manifest-addition",
                                            (List<String>)null );
    return (result == null
        ? new ArrayList<String>()
        : result);
  }


  /**
   * A reference to the {@link Workspace} of which this Project is a part.
   *
   * @return The Project's {@link Workspace}.
   */
  public Workspace workspace()
  {
    return workspace;
  }


  public List<Binaries> getBinaries()
  {
    return binaries;
  }


  public List<Classpath> getClasspath()
  {
    return classPaths;
  }


  public List<Library> getLibraries()
  {
    return libraries;
  }


  public List<ProjectReference> getProjects()
  {
    return projects;
  }


  public List<ProjectReference> getSubProjects()
  {
    if( subProjects == null )
      {
        subProjects = new ArrayList<>();
        processSubdependencies( projects,
                                subProjects );

      }
    return subProjects;
  }


  public List<Sources> getSources()
  {
    return sources;
  }


  // TODO Need ALL other settings for this project (lazy load!)

  @Override
  public String toString()
  {
    return projectName;
  }


  @Override
  public int compareTo( final Project other )
  {
    return projectName.toUpperCase().compareTo( other.projectName.toUpperCase() );
  }


  private void processSubdependencies( final List<ProjectReference> dependentProjects,
                                       final List<ProjectReference> subProjects )
  {
    try
      {
        for( final ProjectReference ref : dependentProjects )
          {
            final Project p = ref.project();
            if( p != null )
              {
                collectDependencies( dependentProjects,
                                     subProjects,
                                     ref.project() );
              }
          }
      }
    catch( final Throwable x )
      {
        x.printStackTrace();
      }
  }


  private void collectDependencies( final List<ProjectReference> dependentProjects,
                                    final List<ProjectReference> subProjects,
                                    final Project project )
  {
    final List<ProjectReference> references = project.getProjects();
    if( references != null )
      {
        for( final ProjectReference ref : references )
          {
            if( !dependentProjects.contains( ref ) && !subProjects.contains( ref ) )
              {
                subProjects.add( ref );
              }
          }
      }
  }


  public void setAdditions( final List<Addition> additions )
  {
    final List<String> values = new ArrayList<>();
    for( final Addition item : additions )
      {
        values.add( Addition.SEP + item.where() + Addition.SEP + item.getActualFile() + Addition.SEP + item.storeAs() );
      }
    config.put( "addition",
                values );
  }


  public List<Addition> getAdditions()
  {
    final List<Addition> result = new ArrayList<>();

    final List<String> values = config.get( "addition",
                                            (List<String>)null );
    if( values != null )
      {
        final Logger logger = Util.logger();
        int n = 0;
        for( final String pattern : values )
          {
            if( pattern.length() > 0 )
              {
                try
                  {
                    final StringTokenizer t = new StringTokenizer( pattern.substring( 1 ),
                                                                   pattern.substring( 0,
                                                                                      1 ) );
                    final Where where = Where.valueOf( t.nextToken() );
                    final File actual = new File( t.nextToken() );
                    final String stored = t.nextToken();

                    result.add( new Addition( actual,
                                              stored,
                                              where ) );
                  }
                catch( final Exception x )
                  {
                    logger.log( Level.SEVERE,
                                "Project " + projectName + " " + "addition." + n + " is bad: " + pattern,
                                x );
                  }
                n++;
              }
          }
      }
    return result;
  }


  public static class Addition
  {
    private static final String SEP = "\u007f";


    public Addition( final File actual,
                     final String stored,
                     final Where where )
    {
      super();
      this.actualFile = actual;
      this.storeAs = stored;
      this.where = where;
    }


    public void setActualFile( final File actualFile )
    {
      this.actualFile = actualFile;
    }


    public File getActualFile()
    {
      return actualFile;
    }


    public void setStoreAs( final String storeAs )
    {
      this.storeAs = storeAs;
    }


    public String storeAs()
    {
      return storeAs;
    }


    public void setWhere( final Where where )
    {
      this.where = where;
    }


    public Where where()
    {
      return where;
    }


    public enum Where
    {
      JAR("ONLY Executable JAR file"),
      ARCHIVE("ONLY distribution archive"),
      BOTH("BOTH executable JAR and Distribution archive");
      private Where( final String toolTipText )
      {
        this.toolTipText = toolTipText;
      }


      public String getToolTipText()
      {
        return toolTipText;
      }


      public synchronized static Where[] textArray()
      {
        if( textArray == null )
          {
            final List<Where> tmp = new ArrayList<>();
            for( final Where w : values() )
              {
                tmp.add( w );
              }
            textArray = new Where[tmp.size()];
            tmp.toArray( textArray );
          }
        return textArray;
      }

      private final String toolTipText;
      private static Where[] textArray;
    }

    private File actualFile;
    private String storeAs;
    private Where where;
  }


  /**
   * <p>
   * The output product of a project, where <code>kind="output"</code>. The
   * object iterates over all files in this group; their relative names (i.e. as
   * referenced in the classpath) are obtained through
   * {@link #toRelativeName(File)}.
   * </p>
   *
   * @author K Udo Schuermann
   */
  public class Binaries
    implements
      Iterable<File>
  {
    public Binaries( final File root )
    {
      super();
      this.root = root;
      this.rootName = root.toString();
      this.rootNameLength = rootName.length() + 1;
    }


    public File root()
    {
      return root;
    }


    @Override
    public Iterator<File> iterator()
    {
      final List<File> result = new ArrayList<>();
      collectFiles( root,
                    result );
      return result.iterator();
    }


    public String toRelativeName( final File file )
    {
      final String name = file.toString();
      return (name.startsWith( rootName )
          ? name.substring( rootNameLength )
          : name);
    }


    private void collectFiles( final File root,
                               final List<File> result )
    {
      final File[] all = root.listFiles();
      Arrays.sort( all );

      // Descend into all directories first
      for( final File file : all )
        {
          if( file.isDirectory() )
            {
              collectFiles( file,
                            result );
            }
        }
      // Process files last
      for( final File file : all )
        {
          if( file.isFile() )
            {
              result.add( file );
            }
        }
    }


    @Override
    public String toString()
    {
      return "Binaries(" + root + ")";
    }

    private final int rootNameLength;
    private final String rootName;
    private final File root;
  }


  /**
   * A directory with classes and other resources to be added to the classpath.
   *
   * @author K Udo Schuermann
   */
  public class Classpath
    implements
      Iterable<File>,
      Comparable<Classpath>
  {
    public Classpath( final File root )
    {
      super();
      this.root = root;
      this.rootName = root.toString();
      this.rootNameLength = rootName.length() + 1;
      this.relativeRoot = relativize( rootName );
    }


    public File root()
    {
      return root;
    }


    public void setExclusions( final List<Exclusion> exclusions )
    {
      if( this.exclusions == null )
        {
          this.exclusions = new ArrayList<>();
        }
      else
        {
          this.exclusions.clear();
        }
      this.exclusions.addAll( exclusions );
      saveExclusions( EXCLUSION_BASEKEY + relativeRoot,
                      exclusions );
    }


    public List<Exclusion> getExclusions()
    {
      if( this.exclusions == null )
        {
          this.exclusions = new ArrayList<>();
          loadExclusions( EXCLUSION_BASEKEY + relativeRoot,
                          this.exclusions );
        }
      final List<Exclusion> result = new ArrayList<>();
      result.addAll( this.exclusions );
      return result;
    }


    @Override
    public int compareTo( final Classpath other )
    {
      return root.compareTo( other.root );
    }


    @Override
    public boolean equals( final Object other )
    {
      if( other instanceof Classpath )
        {
          return (compareTo( (Classpath)other ) == 0);
        }
      return false;
    }


    @Override
    public int hashCode()
    {
      return root.hashCode();
    }


    @Override
    public Iterator<File> iterator()
    {
      if( contents == null )
        {
          contents = new ArrayList<>();
          collectFiles( root,
                        contents );
        }
      return contents.iterator();
    }


    public String toRelativeName( final File file )
    {
      final String name = file.toString();
      return (name.startsWith( rootName )
          ? name.substring( rootNameLength )
          : name);
    }


    private void collectFiles( final File root,
                               final List<File> result )
    {
      final File[] all = root.listFiles();
      Arrays.sort( all );

      // Descend into all directories first
      for( final File file : all )
        {
          if( file.isDirectory() )
            {
              collectFiles( file,
                            result );
            }
        }
      // Process files last
      for( final File file : all )
        {
          if( file.isFile() )
            {
              result.add( file );
            }
        }
    }


    @Override
    public String toString()
    {
      return "ClassPath(" + root + ")";
    }

    private List<Exclusion> exclusions;
    private List<File> contents;
    //
    private final File root;
    private final String rootName;
    private final int rootNameLength;
    private final String relativeRoot;
    private static final String EXCLUSION_BASEKEY = "classpath-exclusion-";
  }


  /**
   * <p>
   * A library (JAR file) of packaged resources. By default, Library files are
   * referenced externally through the Manifest's classpath, but they can also
   * be merged into the main product (to eliminate the volatility of external
   * dependencies and simplify delivery at the cost of a larger main product),
   * or omitted entirely (which is extremely likely to break things, unless the
   * main product is run as a classpath component with the omitted portions also
   * referenced manually: this would be mostly for advanced internal use, or for
   * testing how the software handles a broken external reference).
   * </p>
   *
   * <p>
   * The object iterates over the file contents of the library, and these names
   * can be used to {@linkplain #copyTo(String,OutputStream) transfer that
   * element's data to an OutputStream}.
   * </p>
   *
   * @author K Udo Schuermann
   */
  public class Library
    implements
      Iterable<String>,
      Comparable<Library>
  {
    public Library( final File archive )
    {
      super();
      this.archive = archive;
      this.relativeArchive = relativize( archive.toString() );
    }


    public File archive()
    {
      return archive;
    }


    public void setHandling( final LibraryHandlingOption referenceType )
    {
      config.put( HANDLING_BASEKEY + relativeArchive,
                  referenceType.toString() );
    }


    public LibraryHandlingOption getHandling()
    {
      final String type = config.get( HANDLING_BASEKEY + relativeArchive,
                                      (String)null );
      try
        {
          return LibraryHandlingOption.valueOf( type );
        }
      catch( final Exception x )
        {
          return LibraryHandlingOption.REFERENCE;
        }
    }


    public void setReferencePath( final String path )
    {
      config.put( REFERENCE_BASEKEY + relativeArchive,
                  path );
    }


    public String getReferencePath()
    {
      return config.get( REFERENCE_BASEKEY + relativeArchive,
                         "" );
    }


    @Override
    public Iterator<String> iterator()
    {
      if( contents == null )
        {
          contents = new ArrayList<>();
        }

      try( final ZipFile zipFile = new ZipFile( archive,
                                                ZipFile.OPEN_READ ) )
        {
          final Enumeration<? extends ZipEntry> entries = zipFile.entries();
          while( entries.hasMoreElements() )
            {
              final ZipEntry entry = entries.nextElement();
              if( !entry.isDirectory() )
                {
                  contents.add( entry.getName() );
                }
            }
        }
      catch( final IOException x )
        {
          x.printStackTrace();
        }

      return contents.iterator();
    }


    /**
     * Transfers the associated data to a waiting {@link OutputStream}
     *
     * @param name
     *          The name of the data entry
     * @param output
     *          The OutputStream to receive the data
     * @return 'true' if the data was copied, 'false' if the entry was not
     *         found.
     * @throws IOException
     *           A error occurred reading the archive
     */
    public boolean copyTo( final String name,
                           final OutputStream output )
      throws IOException
    {
      try( final ZipFile zipFile = new ZipFile( archive,
                                                ZipFile.OPEN_READ ) )
        {
          final Enumeration<? extends ZipEntry> entries = zipFile.entries();
          while( entries.hasMoreElements() )
            {
              final ZipEntry entry = entries.nextElement();
              final String entryName = entry.getName();
              if( entryName.equals( name ) )
                {
                  try( final InputStream f = zipFile.getInputStream( entry ) )
                    {
                      Util.copy( new BufferedInputStream( f ),
                                 output );
                    }
                  return true;
                }
            }
        }
      catch( final IOException x )
        {
          x.printStackTrace();
        }
      return false;
    }


    @Override
    public int compareTo( final Library other )
    {
      return archive.compareTo( other.archive );
    }


    @Override
    public boolean equals( final Object other )
    {
      if( other instanceof Library )
        {
          return (compareTo( (Library)other ) == 0);
        }
      return false;
    }


    @Override
    public int hashCode()
    {
      return archive.hashCode();
    }


    @Override
    public String toString()
    {
      return "Library(" + archive + ")";
    }

    private List<String> contents;
    //
    private final File archive;
    private final String relativeArchive;
    private static final String HANDLING_BASEKEY = "library-handling-";
    private static final String REFERENCE_BASEKEY = "library-reference-";
  }


  /**
   * <p>
   * The sources of a project, where <code>kind="src"</code> and the path is a
   * project-relative value. The object iterates over each source; the relative
   * name of the object (as its equivalent binary product would appear in the
   * classpath) can be obtained with {@link #toRelativeName(File)}.
   * </p>
   *
   * @author K Udo Schuermann
   */
  public class Sources
    implements
      Iterable<File>
  {
    public Sources( final File root )
    {
      super();
      this.root = root;
      this.rootName = root.toString();
      this.rootNameLength = rootName.length() + 1;
      this.relativeRoot = relativize( rootName );
    }


    public File root()
    {
      return root;
    }


    public void setHandling( final FileHandlingOption handlingType )
    {
      config.put( HANDLING_BASEKEY + relativeRoot,
                  handlingType.toString() );
    }


    public FileHandlingOption getHandling()
    {
      final String type = config.get( HANDLING_BASEKEY + relativeRoot,
                                      (String)null );
      try
        {
          return FileHandlingOption.valueOf( type );
        }
      catch( final Exception x )
        {
          return FileHandlingOption.IGNORE;
        }
    }


    public void setExclusions( final List<Exclusion> exclusions )
    {
      if( this.exclusions == null )
        {
          this.exclusions = new ArrayList<>();
        }
      else
        {
          this.exclusions.clear();
        }
      this.exclusions.addAll( exclusions );
      saveExclusions( EXCLUSION_BASEKEY + relativeRoot,
                      exclusions );
    }


    public List<Exclusion> getExclusions()
    {
      if( this.exclusions == null )
        {
          this.exclusions = new ArrayList<>();
          loadExclusions( EXCLUSION_BASEKEY + relativeRoot,
                          this.exclusions );
        }
      final List<Exclusion> result = new ArrayList<>();
      result.addAll( this.exclusions );
      return result;
    }


    public void setReferencePath( final String path )
    {
      config.put( REFERENCE_BASEKEY + relativeRoot,
                  path );
    }


    public String getReferencePath()
    {
      return config.get( REFERENCE_BASEKEY + relativeRoot,
                         "" );
    }


    @Override
    public Iterator<File> iterator()
    {
      final List<File> result = new ArrayList<>();
      collectFiles( root,
                    result );
      return result.iterator();
    }


    public String toRelativeName( final File file )
    {
      final String name = file.toString();
      return (name.startsWith( rootName )
          ? name.substring( rootNameLength )
          : name);
    }


    private void collectFiles( final File root,
                               final List<File> result )
    {
      final File[] all = root.listFiles();
      Arrays.sort( all );

      // Descend into all directories first
      for( final File file : all )
        {
          if( file.isDirectory() )
            {
              collectFiles( file,
                            result );
            }
        }
      // Process files last
      for( final File file : all )
        {
          if( file.isFile() )
            {
              result.add( file );
            }
        }
    }


    @Override
    public String toString()
    {
      return "Sources(" + root + ")";
    }

    private List<Exclusion> exclusions;
    //
    private final int rootNameLength;
    private final String rootName;
    private final File root;
    private final String relativeRoot;
    private static final String HANDLING_BASEKEY = "source-handling-";
    private static final String EXCLUSION_BASEKEY = "source-exclusion-";
    private static final String REFERENCE_BASEKEY = "source-reference-";
  }


  /**
   * A reference to a project upon which this project depends.
   *
   * @author K Udo Schuermann
   */
  public class ProjectReference
    implements
      Comparable<ProjectReference>
  {
    public ProjectReference( final Workspace workspace,
                             final String projectName,
                             final boolean isExported,
                             final boolean isCombinedAccessRules )
    {
      super();
      this.workspace = workspace;
      this.isExported = isExported;
      this.isCombinedAccessRules = isCombinedAccessRules;
      this.projectName = projectName;
    }


    public Project project()
    {
      if( project == null )
        {
          project = workspace.findProject( projectName );
        }
      return project;
    }


    public Workspace workspace()
    {
      return workspace;
    }


    public void setHandling( final ArchiveHandlingOption referenceType )
    {
      config.put( HANDLING_BASEKEY + projectName,
                  referenceType.toString() );
    }


    public ArchiveHandlingOption getHandling()
    {
      final String type = config.get( HANDLING_BASEKEY + projectName,
                                      (String)null );
      try
        {
          return ArchiveHandlingOption.valueOf( type );
        }
      catch( final Exception x )
        {
          return ArchiveHandlingOption.REFERENCE;
        }
    }


    public void setReferencePath( final String path )
    {
      config.put( REFERENCE_BASEKEY + projectName,
                  path );
    }


    public String getReferencePath()
    {
      return config.get( REFERENCE_BASEKEY + projectName,
                         "" );
    }


    public boolean isExported()
    {
      return isExported;
    }


    public boolean isCombinedAccessRules()
    {
      return isCombinedAccessRules;
    }


    @Override
    public int compareTo( final ProjectReference other )
    {
      return projectName.compareTo( other.projectName );
    }


    @Override
    public boolean equals( final Object other )
    {
      if( other instanceof ProjectReference )
        {
          return (compareTo( (ProjectReference)other ) == 0);
        }
      return false;
    }


    @Override
    public int hashCode()
    {
      return projectName.hashCode();
    }


    @Override
    public String toString()
    {
      return "Project(" + project() + ")";
    }

    private Project project;
    //
    private final Workspace workspace;
    private final String projectName;
    private final boolean isExported;
    private final boolean isCombinedAccessRules;
    private static final String HANDLING_BASEKEY = "project-handling-";
    private static final String REFERENCE_BASEKEY = "project-reference-";
  }


  private String relativize( final String value )
  {
    final String pRoot = projectRoot.toString();
    if( value.startsWith( pRoot ) )
      {
        return value.substring( pRoot.length() + 1 );
      }

    final String wRoot = workspace().root().toString();
    if( value.startsWith( wRoot ) )
      {
        return value.substring( wRoot.length() );
      }

    return value;
  }


  private void loadExclusions( final String baseKey,
                               final List<Exclusion> exclusions )
  {
    final List<String> exclusionPatterns = config.get( baseKey,
                                                       (List<String>)null );
    if( exclusionPatterns != null )
      {
        final Logger logger = Util.logger();
        int n = 0;
        for( final String pattern : exclusionPatterns )
          {
            try
              {
                final Pattern p = Pattern.compile( pattern );
                exclusions.add( new Exclusion( p ) );
              }
            catch( final PatternSyntaxException x )
              {
                logger.log( Level.SEVERE,
                            "Project " + projectName + " " + baseKey + "." + n + " is bad: " + pattern,
                            x );
              }
            n++;
          }
      }
  }


  private void saveExclusions( final String baseKey,
                               final List<Exclusion> exclusions )
  {
    final List<String> values = new ArrayList<>();
    for( final Exclusion exclusion : exclusions )
      {
        values.add( exclusion.getPattern().toString() );
      }
    config.put( baseKey,
                values );
  }


  /**
   * A {@link DefaultHandler} implementation for an XML SAX parser, which
   * extracts elements from Eclipse's <tt>.classpath</tt> file.
   *
   * @author K Udo Schuermann
   */
  private class ClasspathXmlHandler
    extends DefaultHandler
  {
    /**
     * Constructs a ClasspathHandler for a given Project, and provides a
     * {@link List} within which {@link ClasspathEntry} objects are collected as
     * they are parsed out of the <tt>.classpath</tt> file.
     *
     * @param project
     * @param classpathEntries
     */
    ClasspathXmlHandler()
    {
      super();
    }


    @Override
    public void startElement( final String uri,
                              final String localName,
                              final String qName,
                              final Attributes attributes )
      throws SAXException
    {
      if( inClasspath )
        {
          if( "classpathentry".equals( qName ) )
            {
              final String kind = get( attributes,
                                       "kind" );
              final String path = get( attributes,
                                       "path" );
              final boolean isExported = get( attributes,
                                              "exported",
                                              false );
              final boolean isCombinedAccessRules = get( attributes,
                                                         "combineaccessrules",
                                                         false );

              switch( kind )
                {
                case "src":
                  if( path.startsWith( "/" ) )
                    {
                      // Project reference (within the same workspace)
                      final String name = path.replaceAll( "/",
                                                           "" );
                      projects.add( new ProjectReference( workspace,
                                                          name,
                                                          isExported,
                                                          isCombinedAccessRules ) );
                    }
                  else
                    {
                      // Source reference within the project directory
                      sources.add( new Sources( new File( root(),
                                                          path ) ) );
                    }
                  break;

                case "lib":
                  final File ref = (path.startsWith( "/" )
                      ? new File( path )
                      : new File( root(),
                                  path ));
                  if( ref.isFile() )
                    {
                      libraries.add( new Library( ref ) );
                    }
                  else
                    {
                      classPaths.add( new Classpath( ref ) );
                    }
                  break;

                case "output":
                  binaries.add( new Binaries( new File( root(),
                                                        path ) ) );
                  break;

                case "con":
                  break;
                }
            }
        }
      else if( "classpath".equals( qName ) )
        {
          inClasspath = true;
        }
    }


    @Override
    public void endElement( final String uri,
                            final String localName,
                            final String qName )
    {
      if( "classpathentry".equals( qName ) )
        {
          // nothing needs doing
        }
      else if( "classpath".equals( qName ) )
        {
          inClasspath = false;
        }
    }


    private String get( final Attributes attributes,
                        final String key )
    {
      final int index = attributes.getIndex( key );
      if( index < 0 )
        {
          return null;
        }
      return attributes.getValue( index );
    }


    private boolean get( final Attributes attributes,
                         final String key,
                         final boolean defaultValue )
    {
      final String value = get( attributes,
                                key );
      if( value == null )
        {
          return defaultValue;
        }
      return "true".equalsIgnoreCase( value );
    }

    private boolean inClasspath;
  }

  private boolean mustLoadProjectIcon = true;
  private ImageIcon projectIcon;
  private boolean checkedEntryPoints;
  //
  private final Workspace workspace;
  private final String projectName;
  private final File projectRoot;
  private final Config config;
  private List<Exclusion> binariesExclusions;
  private List<ProjectReference> subProjects;
  private final List<ProjectReference> projects = new ArrayList<>();
  private final List<Sources> sources = new ArrayList<>();
  private final List<Library> libraries = new ArrayList<>();
  private final List<Classpath> classPaths = new ArrayList<>();
  private final List<Binaries> binaries = new ArrayList<>();
  private static final Charset UTF8 = Charset.forName( "UTF-8" );
  private static final int PUBLIC_STATIC = Modifier.PUBLIC | Modifier.STATIC;
  private static final String defaultJNLP = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
                                            + "<!-- JNLP File a hypothetical Demo Application -->\n"
                                            + "<jnlp\n"
                                            + "  spec=\"6.0+\"\n"
                                            + "  codebase=\"http://example.com/downloads/\"\n"
                                            + "  href=\"Demo.jnlp\">\n"
                                            + "      <!-- No 'os' tag: Applies to all -->\n"
                                            + "  <information>\n"
                                            + "    <title>Demo Application</title>\n"
                                            + "    <vendor>Example Inc.</vendor>\n"
                                            + "    <homepage href=\"docs/help.html\"/>\n"
                                            + "    <description>Demo Application</description>\n"
                                            + "    <description kind=\"short\">A demonstration of how to\n"
                                            + "    set up a JNLP file for a Demo Application</description>\n"
                                            + "        <!-- icon kind: default, splash, shortcut -->\n"
                                            + "    <icon href=\"images/demo.jpeg\"/>\n"
                                            + "    <icon kind=\"splash\" href=\"images/splash.png\"/>\n"
                                            + "    <offline-allowed/>\n"
                                            + "    <association mime-type=\"application-x/demo-file\" extensions=\"demo-file\"/> \n"
                                            + "    <shortcut online=\"false\" install=\"false\">\n"
                                            + "      <desktop/>\n"
                                            + "      <menu submenu=\"Example Company Apps\"/>\n"
                                            + "    </shortcut>\n"
                                            + "  </information>\n"
                                            + "  <information os=\"linux\">\n"
                                            + "    <title>Demo Running under Linux</title>\n"
                                            + "    <homepage href=\"docs/linuxhelp.html\">\n"
                                            + "  </information>\n"
                                            + "  <information os=\"mac os x\">\n"
                                            + "    <title>Demo Running under Mac OS/X</title>\n"
                                            + "    <homepage href=\"docs/osxhelp.html\">\n"
                                            + "  </information>\n"
                                            + "  <information os=\"windows\">\n"
                                            + "    <title>Demo Running under Mac OS/X</title>\n"
                                            + "    <homepage href=\"docs/osxhelp.html\">\n"
                                            + "  </information>\n"
                                            + "  <security>\n"
                                            + "    <all-permissions/>\n"
                                            + "  </security>\n"
                                            + "  <resources>\n"
                                            + "    <j2se version=\"1.7+\" java-vm-args=\"-esa -Xnoclassgc\"/>\n"
                                            + "    <jar href=\"Demo.jar\"/>\n"
                                            + "  </resources>\n"
                                            + "  <application-desc main-class=\"com.example.demo.Main\">\n"
                                            + "    <argument>--demo=yes</argument>\n"
                                            + "    <argument>--verbose</argument>\n"
                                            + "    <argument>-X</argument>\n"
                                            + "  </application-desc>\n"
                                            + "</jnlp>";
}