package com.ringlord.boxer2;

import java.io.BufferedOutputStream;
import java.io.Console;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Properties;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import com.ringlord.crypto.PBKDF2;
import com.ringlord.mime.Base64;


/**
 * <p>
 * Manages the contents of the Signing Properties File, whose purpose is to
 * store the credentials needed to digitally sign the application JAR. This
 * Signing Properties File is intended not to be part of the project being built
 * (i.e. not checked into the repository!!) so that each developer has his/her
 * own keystore and credentials, distinct from every other.
 * </p>
 *
 * <p>
 * One limitation is that the name of the Signing Properties File is part of the
 * Boxer configuration file (.boxer2-conf), so every developer would have to
 * store this file in the same place.
 * </p>
 *
 * @author K Udo Schuermann
 */
public class SigningProperties
{
  public static void resetInteractiveTime()
  {
    interactiveTime = 0;
  }


  public static long getInteractiveTime()
  {
    return interactiveTime;
  }


  /**
   * Attempt to load the given Keystore Signing Properties file, interactively
   * asking for a password, either via GUI or on the console.
   *
   * @param file
   *          The Keystore Signing Properties file to load; this file is
   *          expected to contain at least four properties: keystore-file,
   *          keystore-pass, key-alias, and key-pass. Additional properties may
   *          be present to support cryptographic functionality. Note that the
   *          file does not have to exist for it to be "loaded" (it will load
   *          empty, but can then be saved).
   * @param passwordReader
   *          An implementation of PasswordReader that provides for a way to
   *          obtain a password (generally interactively). The Console is used
   *          when this value is null, so the PasswordReader is intended for GUI
   *          operations.
   * @return The KeystoreInfo object, or null if the file could not be loaded
   *         because the proper password did not exist.
   *
   */
  public static SigningProperties load( final File file,
                                        final PasswordReader passwordReader )
  {
    if( passwordReader == null )
      {
        return loadCLI( file );
      }
    return loadGUI( file,
                    passwordReader );
  }


  private static SigningProperties loadGUI( final File file,
                                            final PasswordReader passwordReader )
  {
    String password = null;
    SigningProperties ksInfo;

    while( true )
      {
        try
          {
            ksInfo = SigningProperties.load( file,
                                             password );
            break;
          }
        catch( final SigningProperties.PasswordRequiredException x )
          {
            final long startTime = System.currentTimeMillis();
            password = passwordReader.readPassword();
            interactiveTime += (System.currentTimeMillis() - startTime);
            if( password == null )
              {
                // user canceled
                return null;
              }
          }
        catch( final NoClassDefFoundError x )
          {
            x.printStackTrace();
            System.exit( 0 );
          }
      }
    return ksInfo;
  }


  private static SigningProperties loadCLI( final File file )
  {
    String password = null;
    SigningProperties ksInfo;
    Console console = null;

    while( true )
      {
        try
          {
            ksInfo = SigningProperties.load( file,
                                             password );
            if( password != null )
              {
                System.out.println( "OK" );
              }
            break;
          }
        catch( final SigningProperties.PasswordRequiredException x )
          {
            if( console == null )
              {
                console = System.console();
                if( console == null )
                  {
                    System.err.println( "No Console available, cannot ask for password!" );
                    return null;
                  }
              }
            else
              {
                System.out.println( "Password incorrect!" );
              }
            final long startTime = System.currentTimeMillis();
            // \u0007 is an ASCII BEL (bell/beep)
            final char[] pwd = console.readPassword( "\u0007%s",
                                                     "Password: " );
            interactiveTime += (System.currentTimeMillis() - startTime);
            if( pwd.length == 0 )
              {
                return null;
              }
            password = new String( pwd );
          }
        catch( final NoClassDefFoundError x )
          {
            x.printStackTrace();
            System.exit( 0 );
          }
      }
    return ksInfo;
  }


  /**
   * <p>
   * Construct a Keystore Info object based on a file (existing or not), whose
   * sensitive information (keystore and key passwords) are not expected to be
   * encrypted.
   * </p>
   *
   * @param file
   *          The name of the Signing Properties file whose contents are to be
   *          loaded.
   * @throws PasswordRequiredException
   *           The file's passwords (keystore and key) are encrypted, and a
   *           password is required to access the information properly. Once a
   *           password is obtained (from the user), call the
   *           {@link #KeystoreInfo(File, String) other constructor}, instead.
   */
  public static SigningProperties load( final File file )
    throws PasswordRequiredException
  {
    return load( file,
                 (String)null );
  }


  /**
   * <p>
   * Construct a Keystore Info object based on a file (existing or not) and a
   * password to be used if the file's sensitive information (keystore and key
   * passwords) are encrypted.
   * </p>
   *
   * @param file
   * @param password
   * @throws PasswordRequiredException
   *           The file's passwords (keystore and key) are encrypted, and a
   *           password is required to access the information properly (but no
   *           password (i.e. null) was provided. Once a password is obtained
   *           (from the user), call the {@link #KeystoreInfo(File, String)
   *           other constructor}, instead.
   */
  public static SigningProperties load( final File file,
                                        final String password )
    throws PasswordRequiredException
  {
    final Properties p = new Properties();
    try( final InputStream f = new FileInputStream( file ) )
      {
        p.load( f );
      }
    catch( final IOException x )
      {
      }

    Cipher cipher = null;
    byte[] initVector = null;

    SecretKeySpec key = null;
    byte[] salt = null;

    // Encryption info stores "Algorithm,BASE64(InitVector),BASE64(Salt),PBKDF2"
    // The InitVector and Salt are used to decrypt; new ones are generated each
    // time the information is saved. The 'PBKDF2' is a constant meant to inform
    // casual readers that the user-given password is strengthened into a decent
    // cryptographic key using the PBKDF2 algorithm.
    final String encryptionInfo = p.getProperty( "encryption-info",
                                                 "" ).trim();
    if( encryptionInfo.length() > 0 )
      {
        // Use the encryption info to decrypt stuff, but don't bother keeping
        // the info (only the plain text stuff).

        // Note: In the interest of full disclosure, we are storing this
        // sensitive info in memory for many seconds, potential several minutes,
        // and make no effort to keep the information from proliferating into
        // JPasswordField objects in the GUI, and potentially elsewhere. For
        // this reason there is a chance (albeit small) that an attacker could
        // snoop the process memory and discover the passwords.

        final int comma1 = encryptionInfo.indexOf( "," );
        final int comma2 = encryptionInfo.indexOf( ",",
                                                   comma1 + 1 );
        final int comma3 = encryptionInfo.indexOf( ",",
                                                   comma2 + 1 );
        if( comma3 < 0 )
          {
            throw new IllegalArgumentException( "The file contents seem encrypted, " +
                                                "but some required information is " +
                                                "missing or malformed: " +
                                                file );
          }
        if( comma2 > -1 ) // do we have at least 3 parts?
          {
            if( password == null )
              {
                throw new PasswordRequiredException( "Cannot access encrypted contents of file without password: " +
                                                     file );
              }

            final String algorithm = encryptionInfo.substring( 0,
                                                               comma1 ).trim();
            initVector = Base64.decode( encryptionInfo.substring( comma1 + 1,
                                                                  comma2 ).getBytes() );
            salt = Base64.decode( encryptionInfo.substring( comma2 + 1,
                                                            comma3 ).getBytes() );
            final String pbkdf2 = encryptionInfo.substring( comma3 + 1 );
            if( !"PBKDF2".equals( pbkdf2 ) )
              {
                throw new IllegalArgumentException( "The file contents seem encrypted, " +
                                                    "but some required information is " +
                                                    "missing or malformed: " +
                                                    file );
              }

            try
              {
                int derivedKeySize;
                switch( algorithm )
                  {
                  case "AES":
                  case "AES/CBC/NoPadding":
                    // 128-bit AES
                    cipher = Cipher.getInstance( "AES/CBC/NoPadding" );
                    derivedKeySize = 32;
                    break;
                  case "Blowfish":
                  case "Blowfish/CFB/NoPadding":
                    // 128-bit Blowfish
                    cipher = Cipher.getInstance( "Blowfish/CFB/NoPadding" );
                    derivedKeySize = 16;
                    break;
                  default:
                    cipher = Cipher.getInstance( algorithm );
                    derivedKeySize = 32;
                  }

                final byte[] derivedKey = PBKDF2.deriveKey( password.getBytes( utf8 ),
                                                            salt,
                                                            1024,
                                                            derivedKeySize );
                final int firstSlash = cipher.getAlgorithm().indexOf( "/" );
                final String alg = (firstSlash < 0
                    ? cipher.getAlgorithm()
                    : cipher.getAlgorithm().substring( 0,
                                                       firstSlash ));

                key = new SecretKeySpec( derivedKey,
                                         alg );
              }
            catch( InvalidKeyException |
                   NoSuchAlgorithmException |
                   NoSuchPaddingException e )
              {
                final IllegalArgumentException x = new IllegalArgumentException( "The file contents seem encrypted, " +
                                                                                 "but some required information is " +
                                                                                 "missing or malformed: " +
                                                                                 file );
                x.initCause( e );
                throw x;
              }
          }
      }

    final String keystoreFilename = p.getProperty( "keystore-file",
                                                   "~/.keystore" );
    final String ksPwd = p.getProperty( "keystore-pass" );
    final String keystorePassword;
    if( (cipher != null) && (ksPwd != null) )
      {
        byte[] pwd;
        try
          {
            pwd = decrypt( Base64.decode( ksPwd.getBytes() ),
                           cipher,
                           initVector,
                           key );
            keystorePassword = new String( pwd,
                                           utf8 );
          }
        catch( InvalidKeyException |
               InvalidAlgorithmParameterException |
               IllegalBlockSizeException |
               BadPaddingException e )
          {
            final IllegalArgumentException x = new IllegalArgumentException( "The file contents seem encrypted, " +
                                                                             "but some required information is " +
                                                                             "missing or malformed: " +
                                                                             file );
            x.initCause( e );
            throw x;
          }
      }
    else
      {
        keystorePassword = ksPwd;
      }
    final String keyAlias = p.getProperty( "key-alias",
                                           System.getProperty( "user.name" ) );
    final String keyPwd = p.getProperty( "key-pass" );
    final String keyPassword;
    if( (cipher != null) && (keyPwd != null) )
      {
        try
          {
            final byte[] pwd = decrypt( Base64.decode( keyPwd.getBytes() ),
                                        cipher,
                                        initVector,
                                        key );
            keyPassword = new String( pwd,
                                      utf8 );
          }
        catch( InvalidKeyException |
               InvalidAlgorithmParameterException |
               IllegalBlockSizeException |
               BadPaddingException e )
          {
            final IllegalArgumentException x = new IllegalArgumentException( "The file contents seem encrypted, " +
                                                                             "but some required information is " +
                                                                             "missing or malformed: " +
                                                                             file );
            x.initCause( e );
            throw x;
          }
      }
    else
      {
        keyPassword = keyPwd;
      }

    if( cipher != null )
      {
        try
          {
            final String digestSource = keystorePassword + "\n" + keyPassword;
            MessageDigest sha256;
            sha256 = MessageDigest.getInstance( "SHA-256" );
            final byte[] digest = sha256.digest( digestSource.getBytes( utf8 ) );
            final String digestString = new String( Base64.encode( digest ) );
            if( !digestString.equals( p.getProperty( "encryption-checksum" ) ) )
              {
                throw new PasswordRequiredException( "Wrong password" );
              }
          }
        catch( final NoSuchAlgorithmException e )
          {
            e.printStackTrace();
          }
      }

    final SigningProperties ksInfo = new SigningProperties( p,
                                                            keystoreFilename,
                                                            keystorePassword,
                                                            keyAlias,
                                                            keyPassword,
                                                            cipher );
    ksInfo.password = password;
    return ksInfo;
  }


  public String getKeystoreFilename()
  {
    return keystoreFilename;
  }


  public void setKeystoreFilename( final String keystoreFilename )
  {
    this.keystoreFilename = keystoreFilename;
  }


  public String getKeystorePassword()
  {
    return keystorePassword;
  }


  public void setKeystorePassword( final String keystorePassword )
  {
    this.keystorePassword = keystorePassword;
  }


  public String getKeyAlias()
  {
    return keyAlias;
  }


  public void setKeyAlias( final String keyAlias )
  {
    this.keyAlias = keyAlias;
  }


  public String getKeyPassword()
  {
    return keyPassword;
  }


  public void setKeyPassword( final String keyPassword )
  {
    this.keyPassword = keyPassword;
  }


  public Cipher getCipher()
  {
    return cipher;
  }


  public void setCipher( final Cipher cipher )
  {
    this.cipher = cipher;
  }


  /**
   * @return The password used to decrypt the file contents, if any.
   */
  public String getPassword()
  {
    return password;
  }


  /**
   * Obtains the {@link Properties} object used to load the initial values. This
   * same object is updated and used by the {@link #save} method to ensure that
   * properties otherwise ignored by the KeystoreInfo object are not lost when
   * the information is saved back to disk.
   *
   * @return The Properties from which initial settings were loaded and which
   *         may contain properties that we don't want to lose when saving it
   *         again.
   */
  private Properties getProperties()
  {
    return properties;
  }


  public void save( final File file,
                    final String algorithm,
                    final String password )
    throws IOException
  {
    final Properties p = getProperties();

    if( (algorithm != null) && (password != null) )
      {
        try
          {
            Cipher cipher;
            int derivedKeySize;
            int ivSize;
            switch( algorithm )
              {
              case "AES":
              case "AES/CBC/NoPadding":
                // 128-bit AES
                cipher = Cipher.getInstance( "AES/CBC/NoPadding" );
                derivedKeySize = 32;
                ivSize = 16;
                break;
              case "Blowfish":
              case "Blowfish/CFB/NoPadding":
                // 128-bit Blowfish
                cipher = Cipher.getInstance( "Blowfish/CFB/NoPadding" );
                derivedKeySize = 16;
                ivSize = 8;
                break;
              default:
                cipher = Cipher.getInstance( algorithm );
                derivedKeySize = 32;
                ivSize = 16;
              }

            final byte[] initVector = randomData( new byte[ivSize] );
            final byte[] salt = randomData( new byte[16] );

            final byte[] derivedKey = PBKDF2.deriveKey( password.getBytes( utf8 ),
                                                        salt,
                                                        1024,
                                                        derivedKeySize );

            final int firstSlash = cipher.getAlgorithm().indexOf( "/" );
            final String alg = (firstSlash < 0
                ? cipher.getAlgorithm()
                : cipher.getAlgorithm().substring( 0,
                                                   firstSlash ));
            final SecretKeySpec key = new SecretKeySpec( derivedKey,
                                                         alg );

            final String digestSource = getKeystorePassword() + "\n" + getKeyPassword();
            final MessageDigest sha256 = MessageDigest.getInstance( "SHA-256" );
            final byte[] digest = sha256.digest( digestSource.getBytes( utf8 ) );
            p.put( "encryption-checksum",
                   new String( Base64.encode( digest ) ) );

            p.put( "keystore-pass",
                   new String( Base64.encode( encrypt( getKeystorePassword().getBytes( utf8 ),
                                                       cipher,
                                                       initVector,
                                                       key ) ) ) );
            p.put( "key-pass",
                   new String( Base64.encode( encrypt( getKeyPassword().getBytes( utf8 ),
                                                       cipher,
                                                       initVector,
                                                       key ) ) ) );
            p.put( "encryption-info",
                   cipher.getAlgorithm() +
                       "," +
                       new String( Base64.encode( initVector ) ) +
                       "," +
                       new String( Base64.encode( salt ) ) +
                       "," +
                       "PBKDF2" );
          }
        catch( NoSuchAlgorithmException |
               NoSuchPaddingException |
               InvalidKeyException |
               InvalidAlgorithmParameterException |
               IllegalBlockSizeException |
               BadPaddingException e )
          {
            final IllegalArgumentException x = new IllegalArgumentException( "The file contents seem encrypted, " +
                                                                             "but some required information is " +
                                                                             "missing or malformed: " +
                                                                             file );
            x.initCause( e );
            throw x;
          }
      }
    else
      {
        // Ensure that any old encryption info is removed so that we don't save
        // it, and up on reloading mistake the data as being encrypted.
        p.remove( "encryption-info" );
        p.remove( "encryption-checksum" );

        p.put( "keystore-pass",
               getKeystorePassword() );
        p.put( "key-pass",
               getKeyPassword() );
      }

    p.put( "keystore-file",
           getKeystoreFilename() );
    p.put( "key-alias",
           getKeyAlias() );

    try( final BufferedOutputStream f = new BufferedOutputStream( new FileOutputStream( file ) ) )
      {
        p.store( f,
                 "Saved by Boxer " + Util.VERSION + ", edit with caution, do not distribute!" );
      }
  }


  // TODO I've no idea if Math.random really returns random data
  //
  // NOTE: I distrust SHA1PRNG at this stage, as the NSA has essentially fucked
  // (yes, let my anger shine through!) the security community and made it
  // virtually impossible to trust ANY security-related hardware and software
  // because the NSA may have compromised anything and everything.
  private static byte[] randomData( final byte[] data )
  {
    for( int i = 0; i < data.length; i++ )
      {
        data[i] = (byte)((Math.random() * 256) - 128);
      }
    return data;
  }


  private SigningProperties( final Properties properties,
                             final String keystoreFilename,
                             final String keystorePassword,
                             final String keyAlias,
                             final String keyPassword,
                             final Cipher cipher )
  {
    super();
    this.properties = properties;
    this.keystoreFilename = keystoreFilename;
    this.setKeystorePassword( keystorePassword );
    this.setKeyAlias( keyAlias );
    this.setKeyPassword( keyPassword );
    this.setCipher( cipher );
  }


  private static byte[] decrypt( final byte[] cipherText,
                                 final Cipher cipher,
                                 final byte[] initVector,
                                 final Key key )
    throws InvalidKeyException,
      InvalidAlgorithmParameterException,
      IllegalBlockSizeException,
      BadPaddingException
  {
    final IvParameterSpec iv = new IvParameterSpec( initVector );
    cipher.init( Cipher.DECRYPT_MODE,
                 key,
                 iv );
    final byte[] plainText = cipher.doFinal( cipherText );
    return plainText;
  }


  private static byte[] encrypt( final byte[] plainText,
                                 final Cipher cipher,
                                 final byte[] initVector,
                                 final Key key )
    throws InvalidKeyException,
      InvalidAlgorithmParameterException,
      IllegalBlockSizeException,
      BadPaddingException
  {
    final IvParameterSpec iv = new IvParameterSpec( initVector );
    cipher.init( Cipher.ENCRYPT_MODE,
                 key,
                 iv );
    final byte[] cipherText = cipher.doFinal( plainText );
    return cipherText;
  }


  public interface PasswordReader
  {
    public String readPassword();
  }


  static class PasswordRequiredException
    extends Exception
  {
    PasswordRequiredException( final String message )
    {
      super( message );
    }
    private static final long serialVersionUID = 8734151065268930096L;
  }

  private final Properties properties;
  //
  private String password;
  //
  private String keystoreFilename;
  private String keystorePassword;
  private String keyAlias;
  private String keyPassword;
  private Cipher cipher;
  //
  private static long interactiveTime;
  private static final Charset utf8 = Charset.forName( "UTF-8" );
}
