package com.ringlord.boxer2;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.Security;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import com.ringlord.Filename;
import com.ringlord.LogFormatter;
import com.ringlord.Settings;
import com.ringlord.boxer2.Project.Sources;


/**
 * A set of static utility methods for Boxer, including the {@link Logger},
 * {@link global Preferences}, and the {@link #VERSION}, {@link #VERDATE}, and
 * {@link CPYEARS} values displayed in various places.
 *
 * @author K Udo Schuermann
 */
public class Util
{
  public static final String VERSION = "2.0";
  public static final String VERDATE = "2014-04-26";
  public static final String CPYEARS = "2013,2014";


  public static File resolveFilename( final Project project,
                                      final String filename )
  {
    final String name = resolve( project,
                                 filename );
    return Filename.resolveTilde( name );
  }


  public static String resolve( final Project project,
                                final String value )
  {
    final String sourceFilename = project.getVersionExtractFilename();
    final List<String> extractRegexes = project.getVersionExtractRegexes();
    final List<String> assemblySpecs = project.getVersionExtractAssemblies();
    final String version = getVersion( project,
                                       sourceFilename,
                                       (extractRegexes.size() == 0
                                           ? null
                                           : extractRegexes.get( 0 )),
                                       (assemblySpecs.size() == 0
                                           ? null
                                           : assemblySpecs.get( 0 )) );

    String name;
    if( value.indexOf( "${" ) > -1 )
      {
        name = value;
        name = subst( name,
                      "${VERSION}",
                      (version == null
                          ? ""
                          : version) );
        name = subst( name,
                      "${BUILD}",
                      String.valueOf( project.currentBuildNumber() ) );
        name = subst( name,
                      "${USER}",
                      System.getProperty( "user.name" ) );
        name = subst( name,
                      "${OS}",
                      System.getProperty( "os.name" ) );
        name = subst( name,
                      "${NOW}",
                      new Date().toString() );
      }
    else
      {
        name = value;
      }

    return name;
  }


  private static String subst( String s,
                               final String key,
                               final String value )
  {
    int n;
    while( (n = s.indexOf( key )) > -1 )
      {
        s = s.substring( 0,
                         n ) + value + s.substring( n + key.length() );
      }
    return s;
  }


  public static void copy( final File from,
                           final OutputStream to )
    throws IOException
  {
    try( final InputStream f = new FileInputStream( from ) )
      {
        copy( f,
              to );
      }
    to.flush();
  }


  public static void copy( final InputStream from,
                           final OutputStream to )
    throws IOException
  {
    final byte[] buffer = new byte[4096];
    copy( from,
          to,
          buffer );
  }


  public static void copy( final InputStream from,
                           final OutputStream to,
                           final byte[] buffer )
    throws IOException
  {
    int inBuffer;
    while( (inBuffer = from.read( buffer )) > -1 )
      {
        to.write( buffer,
                  0,
                  inBuffer );
      }
    to.flush();
  }


  public static String getVersion( final Project project,
                                   final String sourceFilename,
                                   final String extractRegex,
                                   final String assemblySpec )
  {
    try
      {
        // If the regex is junk, fail before we parse the file
        final Pattern extract = Pattern.compile( ".*" + extractRegex + ".*",
                                                 Pattern.DOTALL );

        File sourceFile = null;
        if( sourceFilename.startsWith( "/" ) )
          {
            sourceFile = new File( project.workspace().root(),
                                   sourceFilename.substring( 1 ) );
          }
        else
          {
            final String src = sourceFilename.substring( 1 );
            for( final Sources source : project.getSources() )
              {
                final File f = new File( source.root(),
                                         src );
                if( f.exists() )
                  {
                    sourceFile = f;
                    break;
                  }
              }
          }

        return Util.getVersion( sourceFile,
                                extract,
                                assemblySpec );
      }
    catch( final Exception x )
      {
        return null;
      }
  }


  public static String getVersion( final File source,
                                   final Pattern extract,
                                   final String assembly )
  {
    final StringBuilder fileContent = new StringBuilder();
    try( final BufferedInputStream f = new BufferedInputStream( new FileInputStream( source ) ) )
      {
        final byte[] buffer = new byte[1024];
        int inBuffer;
        while( (inBuffer = f.read( buffer )) > -1 )
          {
            if( inBuffer > 0 )
              {
                fileContent.append( new String( buffer,
                                                0,
                                                inBuffer ) );
              }
          }

        final String[] groups;
        final Matcher m = extract.matcher( fileContent );
        if( m.matches() )
          {
            final int groupCount = m.groupCount();
            groups = new String[groupCount];
            for( int i = 0; i < groupCount; i++ )
              {
                groups[i] = m.group( i + 1 );
              }
          }
        else
          {
            groups = new String[0];
          }

        final StringBuilder sb = new StringBuilder();
        final int len = assembly.length();
        boolean isEscaped = false;
        int state = 0;
        int startIndex = 0;
        for( int i = 0; i < len; i++ )
          {
            final char c = assembly.charAt( i );
            if( isEscaped )
              {
                sb.append( c );
                isEscaped = false;
              }
            else
              {
                switch( state )
                  {
                  case 0:
                    if( c == '\\' )
                      {
                        isEscaped = true;
                      }
                    else if( c == '$' )
                      {
                        state = 1;
                      }
                    else
                      {
                        sb.append( c );
                      }
                    break;

                  case 1:
                    if( c == '(' )
                      {
                        state = 2;
                        startIndex = i + 1;
                      }
                    else
                      {
                        sb.append( '$' ).append( c );
                        state = 0;
                      }
                    break;

                  case 2:
                    if( c == ')' )
                      {
                        final String indexString = assembly.substring( startIndex,
                                                                       i );
                        try
                          {
                            final int index = Integer.parseInt( indexString );
                            sb.append( groups[index - 1] );
                          }
                        catch( NumberFormatException |
                               ArrayIndexOutOfBoundsException x )
                          {
                            sb.append( "$(" ).append( indexString ).append( c );
                          }
                        state = 0;
                      }
                    break;
                  }
              }
          }
        return sb.toString();
      }
    catch( final IOException x )
      {
        return null;
      }
  }


  public static synchronized Logger logger()
  {
    if( logger == null )
      {
        // Setup the global logger
        logger = Logger.getGlobal();
        logger.setUseParentHandlers( false );
        logger.setLevel( Level.ALL );
        try
          {
            final File logFile = new File( System.getProperty( "java.io.tmpdir" ),
                                           "boxer-%u%g.log" );
            final FileHandler fileHandler = new FileHandler( logFile.toString(),
                                                             Integer.MAX_VALUE,
                                                             10 );
            fileHandler.setLevel( Level.ALL );
            fileHandler.setFormatter( new LogFormatter() );
            logger.addHandler( fileHandler );
          }
        catch( final IOException bummer )
          {
            System.err.println( "Cannot initialize logger: " + bummer );
            logger.log( Level.WARNING,
                        "Cannot initialize logger",
                        bummer );
          }
      }
    return logger;
  }


  public static synchronized Settings getSettings()
  {
    if( preferences == null )
      {
        preferences = Settings.getInstance( Util.class );
      }
    return preferences;
  }


  public static synchronized List<String> getMessageDigestAlgorithms( final Project project )
  {
    final List<String> digestNames = new ArrayList<>();
    final Provider[] providers = Security.getProviders();
    for( int i = 0; i != providers.length; i++ )
      {
        for( final Object entry : providers[i].keySet() )
          {
            final String algorithm = (String)entry;
            final int index;
            if( (index = algorithm.indexOf( "MessageDigest." )) > -1 )
              {
                String item = algorithm.substring( index + "MessageDigest.".length() );
                // Skip items with blank spaces in them as well as ones that
                // specify no numeric bit size.
                if( (item.indexOf( " " ) < 0) && item.matches( ".*[0-9]+.*" ) )
                  {
                    item = item.replaceAll( "-",
                                            "" ); // trim dashes, just in case
                    // avoid duplicates, too
                    if( !digestNames.contains( item ) )
                      {
                        if( index > 0 )
                          {
                            if( legacyNames == null )
                              {
                                legacyNames = new ArrayList<>();
                              }
                            legacyNames.add( item );
                            Util.logger().log( Level.FINEST,
                                               "Unusual (deprecated?) class name for '" + item + "': " + algorithm );
                          }
                        digestNames.add( item );
                        Util.logger().log( Level.CONFIG,
                                           "Recognized available MessageDigest algorithm '" +
                                               item +
                                               "' (" +
                                               algorithm +
                                               ")" );
                      }
                  }
              }
          }
      }

    if( project != null )
      {
        for( final String algorithm : project.getDigestsPicked() )
          {
            if( !digestNames.contains( algorithm ) )
              {
                try
                  {
                    MessageDigest.getInstance( algorithm ); // return value not
                                                            // needed
                    // We could instantiate this algorithm!
                    digestNames.add( algorithm );
                    if( legacyNames == null )
                      {
                        legacyNames = new ArrayList<>();
                      }
                    legacyNames.add( algorithm );
                    Util.logger().log( Level.CONFIG,
                                       "Previously used (legacy?) MessageDigest algorithm '" +
                                           algorithm +
                                           "' appears available but was not indicated to " +
                                           "be available." );
                  }
                catch( final NoSuchAlgorithmException x )
                  {
                    Util.logger().log( Level.CONFIG,
                                       "Previously used (legacy?) MessageDigest algorithm '" +
                                           algorithm +
                                           "' is no longer available!" );
                  }
              }
          }
      }
    return digestNames;
  }


  public static boolean isLegacyMessageDigestAlgorithm( final String algorithm )
  {
    return (legacyNames != null) && legacyNames.contains( algorithm );
  }


  public static void setKnownWorkspaces( final List<Workspace> workspaces )
  {
    final Settings prefs = getSettings();
    final List<String> all = new ArrayList<>();
    for( final Workspace w : workspaces )
      {
        all.add( w.root().toString() );
      }
    prefs.put( "workspaces",
               all,
               null );
  }


  public static List<Workspace> getKnownWorkspaces()
  {
    final List<Workspace> result = new ArrayList<>();

    final Settings prefs = getSettings();
    final List<String> workspaceNames = prefs.get( "workspaces",
                                                   null,
                                                   null );

    if( workspaceNames != null )
      {
        for( final String ws : workspaceNames )
          {
            final File f = new File( ws );
            if( f.isDirectory() )
              {
                try
                  {
                    final Workspace workspace = new Workspace( f );
                    result.add( workspace );
                  }
                catch( ParserConfigurationException |
                       SAXException |
                       IOException x )
                  {
                    logger().log( Level.SEVERE,
                                  "Failed to open workspace  " + f,
                                  x );
                  }
              }
            else
              {
                logger().log( Level.WARNING,
                              "Workspace '" + f + "' is not a directory, was it deleted?" );
              }
          }
      }
    Collections.sort( result );
    return result;
  }


  public static void setCurrentWorkspace( final String name )
  {
    final Settings prefs = getSettings();
    prefs.put( "current-workspace",
               name );
  }


  public static String getCurrentWorkspace()
  {
    final Settings prefs = getSettings();
    return prefs.get( "current-workspace",
                      (String)null );
  }


  /**
   * Locates all projects in all known workspaces.
   *
   * @param projectDirectory
   * @return
   */
  public static List<Project> findProject( final String projectDirectory )
  {
    final List<String> projectNames = new ArrayList<>();
    projectNames.add( projectDirectory );
    return findProjects( projectNames );
  }


  /**
   * Locates projects matching a given set of (partial) names.
   *
   * @param names
   *          Names (or partial) names for the projects being sought. Each name
   *          must resolve uniquely to precisely one project, otherwise an
   *          IllegalArgumentException is thrown. Note that an exact match
   *          trumps any additional partial matches, so a name 'test' will be
   *          okay when the available projects are 'test', 'test1a', and
   *          'test1b', but a name 'test1' will not be a unique match.
   * @return A set of projects, each matching one of the input names.
   * @throws IllegalArgumentException
   *           One or more of the given names did not match precisely one
   *           project. The exception's message is a set of one or more lines
   *           describing all the failures encountered, including no match at
   *           all, and multiple matches for a partial name.
   */
  public static List<Project> findProjects( final List<String> names )
  {
    final List<Project> projects = new ArrayList<>();

    final Settings prefs = getSettings();
    final List<String> workspaceNames = prefs.get( "workspaces",
                                                   null,
                                                   null );

    String NL = null;
    StringBuilder errorMessage = null;
    for( final String name : names )
      {
        final List<Project> result = findProject( workspaceNames,
                                                  name );
        if( result.size() == 1 )
          {
            // We found exactly one match and it is a Project object
            projects.add( result.get( 0 ) );
          }
        else if( result.isEmpty() )
          {
            if( errorMessage == null )
              {
                errorMessage = new StringBuilder();
                NL = System.getProperty( "line.separator" );
                errorMessage.append( "ERROR: Cannot process project(s):" + NL );
              }
            errorMessage.append( "\tNo project matched '" + name + "'" + NL );
          }
        else
          {
            // Ambiguous!
            if( errorMessage == null )
              {
                errorMessage = new StringBuilder();
                NL = System.getProperty( "line.separator" );
                errorMessage.append( "ERROR: Cannot process project(s):" + NL );
              }
            errorMessage.append( "\tMore than one project matched '" + name + "'" + NL );
            for( final Project p : result )
              {
                errorMessage.append( "\t\t" + p.root() + NL );
              }
          }
      }

    if( errorMessage != null )
      {
        throw new IllegalArgumentException( errorMessage.toString() );
      }

    return projects;
  }


  private static List<Project> findProject( final List<String> workspaceNames,
                                            final String projectDirectory )
  {
    final List<Project> candidates = new ArrayList<>();

    File p = new File( projectDirectory );
    File w = p.getParentFile(); // the parent is the workspace
    if( w == null )
      {
        final String ucName = projectDirectory.toUpperCase();
        search: for( final String ws : workspaceNames )
          {
            final File f = new File( ws );
            if( f.isDirectory() )
              {
                try
                  {
                    final Workspace workspace = new Workspace( f );
                    for( final Project project : workspace.projects() )
                      {
                        final String name = project.name();
                        if( name.equals( projectDirectory ) )
                          {
                            // exact match!
                            w = f;
                            p = new File( w,
                                          projectDirectory );
                            candidates.clear();
                            candidates.add( project );
                            break search;
                          }
                        if( name.toUpperCase().indexOf( ucName ) > -1 )
                          {
                            candidates.add( project );
                          }
                      }
                  }
                catch( final ParserConfigurationException |
                             SAXException |
                             IOException x )
                  {
                  }
              }
          }
      }
    return candidates;
  }


  /**
   * A log formatter intended to show only the message text, with exceptions
   * dumped to stderr.
   *
   * @author K Udo Schuermann
   */
  public static final class PlainLog
    extends Formatter
  {
    PlainLog()
    {
      this.EOLN = System.getProperty( "line.separator" );
    }


    @Override
    public String format( final LogRecord record )
    {
      Throwable t = record.getThrown();
      if( t != null )
        {
          final StringBuilder sb = new StringBuilder();
          while( t != null )
            {
              final String msg = t.getMessage();
              sb.append( t.getClass().getName() ).append( ": " ).append( (msg == null
                  ? ""
                  : msg.trim()) ).append( EOLN );
              final StackTraceElement[] trace = t.getStackTrace();
              for( final StackTraceElement e : trace )
                {
                  sb.append( "\tat " ).append( e.getClassName() ).append( '#' ).append( e.getMethodName() );
                  final String filename = e.getFileName();
                  final int fileline = e.getLineNumber();
                  final boolean isNative = e.isNativeMethod();
                  if( filename != null )
                    {
                      sb.append( " (" ).append( filename );
                      if( fileline >= 0 )
                        {
                          sb.append( ":" + fileline );
                        }
                      sb.append( ')' ).append( EOLN );
                    }
                  else if( isNative )
                    {
                      sb.append( " (native)" ).append( EOLN );
                    }
                  else
                    {
                      sb.append( EOLN );
                    }
                }
              if( (t = t.getCause()) != null )
                {
                  sb.append( "Caused by " );
                }
            }
          System.err.println( sb.toString() );
        }

      final Level level = record.getLevel();
      if( level == Level.SEVERE )
        {
          return "ERROR: " + record.getMessage() + EOLN;
        }
      else if( level == Level.WARNING )
        {
          return "WARNING: " + record.getMessage() + EOLN;
        }
      else if( level == Level.CONFIG )
        {
          return ">>> " + record.getMessage() + EOLN;
        }
      else
        {
          return record.getMessage() + EOLN;
        }
    }

    private final String EOLN;
  }

  private static Logger logger;
  private static Settings preferences;
  private static List<String> legacyNames;
}
