package com.ringlord.boxer2;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;


/**
 * The reference to the Eclipse workspace, by default named "workspace" and
 * located in the current user's home directory. The Workspace supplies a
 * configured {@link SAXParser} for parsing each {@link Project}'s
 * <tt>.classpath</tt> file, and provides a list of Projects in the workspace.
 *
 * @author K Udo Schuermann
 */
public class Workspace
  implements
    Comparable<Workspace>
{
  /**
   * Construct a Workspace object whose root is the given file.
   *
   * @param root
   *          The directory of the Workspace; each directory within the
   *          Workspace is a potential {@link Project} (if the project directory
   *          contains a <tt>.classpath</tt> file).
   * @throws ParserConfigurationException
   *           The {@link SAXParser} intended for parsing each {@link Project}'s
   *           <tt>.classpath</tt> file, could not be initialized.
   * @throws SAXException
   *           A parsing exception occurred while parsing a <tt>.classpath</tt>
   *           file in one of the {@link Project}S.
   * @throws IOException
   *           A file could not be processed while building a list of the
   *           {@link Project}S in this Workspace.
   */
  public Workspace( final File root )
    throws ParserConfigurationException,
      SAXException,
      IOException
  {
    super();

    if( !(root.exists() && root.isDirectory()) )
      {
        throw new IOException( "Not an existing workspace directory: " + root );
      }

    this.root = root;
    this.name = Util.getSettings().get( "workspace-name-" + root,
                                        root.getName() );

    final SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
    saxParserFactory.setValidating( true );
    saxParserFactory.setNamespaceAware( true );
    saxParserFactory.setXIncludeAware( true );

    this.saxParser = saxParserFactory.newSAXParser();

    final String[] files = root.list();
    for( final String name : files )
      {
        try
          {
            final Project p = new Project( this,
                                           name );
            projects.add( p );
            projectMap.put( name,
                            p );
          }
        catch( final IOException |
                     SAXException x )
          {
            // Ignore projects that cannot be properly instantiated for lack of
            // .classpath file, an unknown organization, or corruption.
            Util.logger().log( Level.WARNING,
                               "Unusable project: " + name,
                               x );
          }
        catch( final IllegalArgumentException x )
          {
            // No .classpath file was found, so this is not an actual project
            Util.logger().log( Level.CONFIG,
                               x.getMessage() );
          }
      }
    Collections.sort( projects );
  }


  /**
   * Obtains the directory where this Workspace is located.
   *
   * @return The directory of this Workspace.
   */
  public File root()
  {
    return root;
  }


  public void setName( final String name )
  {
    if( (name == null) || (name.length() == 0) || name.equals( root.getName() ) )
      {
        this.name = root.getName();
        Util.getSettings().put( "workspace-name-" + root,
                                (String)null );
      }
    else
      {
        this.name = name;
        Util.getSettings().put( "workspace-name-" + root,
                                name );
      }
  }


  public String getName()
  {
    return name;
  }


  /**
   * Obtains the SAXParser configured to parse the <tt>.classpath</tt> files of
   * each {@link Project}.
   *
   * @return The SAXParser intended for parsing <tt>.classpath</tt> files.
   */
  SAXParser saxParser()
  {
    return saxParser;
  }


  /**
   * Locate the named project, provided that it exists.
   *
   * @param name
   *          The case-sensitive name of the project, equivalent to the name of
   *          its directory within this Workspace
   * @return The {@link Project} object matching the given project name, or null
   *         if no project by this name exists (note that project names are case
   *         sensitive).
   * @see #projects()
   */
  public Project findProject( final String name )
  {
    return projectMap.get( name );
  }


  /**
   * Obtains a set of all known {@link Project}S within this Workspace. This set
   * does not necessarily represent each directory, as some directories in the
   * Workspace may not contain a <tt>.classpath</tt> file, and thus do not look
   * like {@link Project} directories to Boxer.
   *
   * @return A set of zero or more {@link Project}S; the set can be of zero
   *         length, but is never null.
   * @see #findProject(String)
   */
  public List<Project> projects()
  {
    final List<Project> result = new ArrayList<>();
    result.addAll( projects );
    return result;
  }


  @Override
  public String toString()
  {
    final StringBuilder sb = new StringBuilder();
    final String NL = System.getProperty( "line.separator" );
    System.out.println( "Workspace '" + root + "'" );
    for( final Project p : projects() )
      {
        sb.append( NL + "\tProject " + p.name() );
        sb.append( NL + "\t\tPath: " + p.root() );
        sb.append( NL + "\t\tSources:" );
        for( final Project.Sources r : p.getSources() )
          {
            sb.append( NL + "\t\t\t" + r );
          }
        for( final Project.ProjectReference r : p.getProjects() )
          {
            sb.append( NL + "\t\t\t" + r );
          }
        for( final Project.Classpath r : p.getClasspath() )
          {
            sb.append( NL + "\t\t\t" + r );
          }
        for( final Project.Library r : p.getLibraries() )
          {
            sb.append( NL + "\t\t\t" + r );
          }
      }
    return sb.toString();
  }


  @Override
  public int compareTo( final Workspace other )
  {
    return root.compareTo( other.root );
  }


  @Override
  public boolean equals( final Object other )
  {
    if( other instanceof Workspace )
      {
        return (compareTo( (Workspace)other ) == 0);
      }
    return false;
  }


  @Override
  public int hashCode()
  {
    return root.hashCode();
  }

  private final Map<String,Project> projectMap = new HashMap<>();
  private final List<Project> projects = new ArrayList<>();
  private final File root;
  private String name;
  private final SAXParser saxParser;
}
