package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Insets;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Collections;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;

import com.ringlord.boxer2.LibraryHandlingOption;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Library;
import com.ringlord.swing.Form;


public class ConfigLibraries
  extends ConfigurationPanel
{
  public ConfigLibraries( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigLibraries" );

    // ==================================================

    final List<Library> libraries = project.getLibraries();
    Collections.sort( libraries );
    for( final Library library : libraries )
      {
        createEntry( form,
                     project,
                     library );
      }

    /**
     * <pre>
     * // If there happen to be projects referenced by the dependent projects, then
     * // make them available here, too, so that we can control during the build
     * // how those dependencies are processed. This is likely a very rare case,
     * // but we want to handle it for maximum control.
     * final List&lt;Library&gt; subLibraries = project.getSubLibraries();
     * Collections.sort( subLibraries );
     * for( final Library library : subLibraries )
     *   {
     *     createEntry( form,
     *                  project,
     *                  library );
     *   }
     * </pre>
     */

    // ==================================================

    final JTextArea info = new JTextArea( 4,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "Referencing a library is the traditional and modular approach: It is easy to update "
                  + "one library independently of the others, but it can also be more of a challenge "
                  + "to ensure that all the pieces are placed where they are needed. An alternative is "
                  + "to merge a library's contents into your project so that you can deliver a single "
                  + "package. For desktop apps this approach tends to be the best " );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
  }


  private void createEntry( final Form form,
                            final Project project,
                            final Library library )
  {
    final JComboBox<LibraryHandlingOption> handling = new JComboBox<>( LibraryHandlingOption.textArray() );
    handling.setRenderer( renderer );
    handling.setSelectedItem( library.getHandling() );
    handling.setToolTipText( "<html>"
                             + "The traditional way is to REFERENCE an external library<br>"
                             + "via the classpath, but if licensing permits you could MERGE<br>"
                             + "the library into your main archive (this would simplify<br>"
                             + "deployment!); if you choose to IGNORE the library then<br>"
                             + "it is up to you to resolve the resulting runtime dependencies" );

    final JTextField referencePathField = new JTextField( library.getReferencePath() );
    final JLabel referencePathLabel = new JLabel( "Library path:",
                                                  JLabel.TRAILING );
    final JLabel referenceJarLabel = new JLabel( library.archive().getName() );
    referenceJarLabel.setToolTipText( "The referenced project's archive filename" );
    referencePathLabel.setLabelFor( referencePathField );
    setToolTipText( "<html>" +
                        "The path to the referenced project archive,<br>" +
                        "relative to the main archive. This becomes<br>" +
                        "part of the classpath and could be a simple<br>" +
                        "subdirectory like <code>" +
                        project.name() +
                        "-libs/</code>",
                    referencePathLabel,
                    referencePathField );

    final Form filter = new Form();
    filter.addRow( referencePathLabel,
                   referencePathField,
                   referenceJarLabel );

    final String pn = project.root().toString();
    final String ws = project.workspace().root().toString();
    String root = library.archive().toString();
    if( root.startsWith( pn ) )
      {
        root = root.substring( pn.length() );
      }
    else if( root.startsWith( ws ) )
      {
        root = root.substring( ws.length() );
      }

    if( handling.getSelectedItem() != LibraryHandlingOption.REFERENCE )
      {
        filter.setVisible( false );
      }

    handling.addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
        final Object selection = handling.getSelectedItem();
        if( selection == LibraryHandlingOption.IGNORE )
          {
            filter.setVisible( false );
            library.setHandling( LibraryHandlingOption.IGNORE );
          }
        else if( selection == LibraryHandlingOption.MERGE )
          {
            filter.setVisible( false );
            library.setHandling( LibraryHandlingOption.MERGE );
          }
        else if( selection == LibraryHandlingOption.REFERENCE )
          {
            filter.setVisible( true );
            library.setHandling( LibraryHandlingOption.REFERENCE );
          }
        else
          {
            // TODO log this issue
            System.err.println( "Unhandled LibraryHandlingOption: " + selection );
          }
        form.revalidate();
      }
    } );
    final ComponentListener unifiedListener = new ComponentListener()
    {
      @Override
      public void modified( final Object source )
      {
        library.setReferencePath( referencePathField.getText().trim() );
      }
    };
    referencePathField.addActionListener( unifiedListener );
    referencePathField.addFocusListener( unifiedListener );

    form.addRow( new JLabel( root,
                             JLabel.TRAILING ),
                 handling );
    form.addRow( null,
                 filter );
  }


  private class Renderer
    extends JLabel
    implements
      ListCellRenderer<LibraryHandlingOption>
  {
    public Renderer()
    {
      super();
      setOpaque( true );
      setBorder( BorderFactory.createEmptyBorder( 2,
                                                  4,
                                                  2,
                                                  0 ) );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends LibraryHandlingOption> list,
                                                   final LibraryHandlingOption option,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      setText( option.toString() + " \u2014 " + option.getToolTipText() );
      if( isSelected )
        {
          setForeground( list.getSelectionForeground() );
          setBackground( list.getSelectionBackground() );
        }
      else
        {
          setForeground( list.getForeground() );
          setBackground( Color.white );
        }
      return this;
    }

    private static final long serialVersionUID = -6664320180866633649L;
  }

  private final Renderer renderer = new Renderer();
  private static final long serialVersionUID = 8275807297511691099L;
}
