package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Insets;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.util.Collections;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;

import com.ringlord.boxer2.ArchiveHandlingOption;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.ProjectReference;
import com.ringlord.swing.Form;


public class ConfigProjectRefs
  extends ConfigurationPanel
{
  public ConfigProjectRefs( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigProjectRefs" );

    // ==================================================

    final List<ProjectReference> projectRefs = project.getProjects();
    Collections.sort( projectRefs );
    for( final ProjectReference projectRef : projectRefs )
      {
        createEntry( form,
                     project,
                     projectRef );
      }

    // If there happen to be projects referenced by the dependent projects, then
    // make them available here, too, so that we can control during the build
    // how those dependencies are processed. This is likely a very rare case,
    // but we want to handle it for maximum control.
    final List<ProjectReference> subProjectRefs = project.getSubProjects();
    Collections.sort( subProjectRefs );
    for( final ProjectReference projectRef : subProjectRefs )
      {
        createEntry( form,
                     project,
                     projectRef );
      }

    // ==================================================

    final JTextArea info = new JTextArea( 4,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "Referencing another project's JAR-packaged binaries is the traditional and modular "
                  + "approach: It is easy to update one JAR independently of the others, but it can "
                  + "also be more of a challenge to ensure that all the JARs are placed where they "
                  + "are needed. An alternative is to merge the binaries from other projects so that "
                  + "you can deliver a single package. For desktop apps this approach tends to be the "
                  + "best." );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
  }


  private void createEntry( final Form form,
                            final Project project,
                            final ProjectReference projectRef )
  {
    final JComboBox<ArchiveHandlingOption> handling = new JComboBox<>( ArchiveHandlingOption.textArray() );
    handling.setRenderer( renderer );
    handling.setSelectedItem( projectRef.getHandling() );
    handling.setToolTipText( "<html>"
                             + "The traditional way is to REFERENCE a dependent<br>"
                             + "project's archive file as an external resource via the<br>"
                             + "classpath, but if licensing permits you could MERGE<br>"
                             + "the other project into your main archive (this would<br>"
                             + "simplify deployment!); if you choose to IGNORE the<br>"
                             + "dependent project then it is up to you to resolve<br>"
                             + "the resulting runtime dependencies." );

    final JTextField referencePathField = new JTextField( projectRef.getReferencePath() );
    final JLabel referencePathLabel = new JLabel( "Archive path:",
                                                  JLabel.TRAILING );
    final JLabel referenceJarLabel = new JLabel( new File( projectRef.project().getOutputJarFile() ).getName() );
    referenceJarLabel.setToolTipText( "The referenced project's archive filename" );
    referencePathLabel.setLabelFor( referencePathField );
    setToolTipText( "<html>" +
                        "The path to the referenced project archive,<br>" +
                        "relative to the main archive. This becomes<br>" +
                        "part of the classpath and could be a simple<br>" +
                        "subdirectory like <code>" +
                        project.name() +
                        "-libs/</code>",
                    referencePathLabel,
                    referencePathField );

    final Form filter = new Form();
    filter.setName( "ConfigProjectRefs." + projectRef.project().name() );
    filter.addRow( referencePathLabel,
                   referencePathField,
                   referenceJarLabel );

    final String ws = project.workspace().root().toString();
    String root = projectRef.project().root().toString();
    if( root.startsWith( ws ) )
      {
        root = root.substring( ws.length() );
      }

    if( handling.getSelectedItem() != ArchiveHandlingOption.REFERENCE )
      {
        filter.setVisible( false );
      }

    handling.addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
        final Object selection = handling.getSelectedItem();
        if( selection == ArchiveHandlingOption.IGNORE )
          {
            filter.setVisible( false );
            projectRef.setHandling( ArchiveHandlingOption.IGNORE );
          }
        else if( selection == ArchiveHandlingOption.MERGE )
          {
            filter.setVisible( false );
            projectRef.setHandling( ArchiveHandlingOption.MERGE );
          }
        else if( selection == ArchiveHandlingOption.REFERENCE )
          {
            filter.setVisible( true );
            projectRef.setHandling( ArchiveHandlingOption.REFERENCE );
          }
        else
          {
            // TODO log this issue
            System.err.println( "Unhandled ArchiveHandlingOption: " + selection );
          }
        form.revalidate();
      }
    } );
    final ComponentListener unifiedListener = new ComponentListener()
    {
      @Override
      public void modified( final Object source )
      {
        projectRef.setReferencePath( referencePathField.getText().trim() );
      }
    };
    referencePathField.addActionListener( unifiedListener );
    referencePathField.addFocusListener( unifiedListener );

    form.addRow( new JLabel( root,
                             JLabel.TRAILING ),
                 handling );
    form.addRow( null,
                 filter );
  }


  private static class Renderer
    extends JLabel
    implements
      ListCellRenderer<ArchiveHandlingOption>
  {
    public Renderer()
    {
      super();
      setOpaque( true );
      setBorder( BorderFactory.createEmptyBorder( 2,
                                                  4,
                                                  2,
                                                  0 ) );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends ArchiveHandlingOption> list,
                                                   final ArchiveHandlingOption option,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      setText( option.toString() + " \u2014 " + option.getToolTipText() );
      if( isSelected )
        {
          setForeground( list.getSelectionForeground() );
          setBackground( list.getSelectionBackground() );
        }
      else
        {
          setForeground( list.getForeground() );
          setBackground( Color.white );
        }
      return this;
    }

    private static final long serialVersionUID = 3086727156758833569L;
  }

  private static final Renderer renderer = new Renderer();
  private static final long serialVersionUID = 8275807297511691099L;
}
