package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.LayoutManager2;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;

import com.ringlord.swing.Form;
import com.ringlord.swing.Images;


/**
 * A JPanel with some extra convenience features to make it a good basis for
 * Config panels.
 *
 * @author K Udo Schuermann
 */
public abstract class ConfigurationPanel
  extends JPanel
{
  protected ConfigurationPanel()
  {
    this( new BorderLayout() );
  }


  protected ConfigurationPanel( final LayoutManager2 layout )
  {
    super( layout );
  }


  protected void setToolTipText( final String toolTipText,
                                 final JComponent ... components )
  {
    for( final JComponent c : components )
      {
        c.setToolTipText( toolTipText );
      }
  }


  /**
   * <p>
   * An generic editor for 0+ rows of 1+ components; any row can be removed, and
   * new rows appended. This is used to define which sources, libraries,
   * additions, etc. are to be excluded, transformed, etc.
   * </p>
   *
   * <p>
   * Start by defining an object that implements RowEditor.Model: This object
   * maintains the values and produced a {@link Component} for any value on a
   * row and column. You must monitor changes to these components and update
   * your data model accordingly.
   * </p>
   *
   * @author K Udo Schuermann
   */
  public static class RowEditor
    extends JPanel
  {
    public RowEditor( final Model model )
    {
      this( model,
            null );
    }


    public RowEditor( final Model model,
                      final String addButtonText )
    {
      super( new BorderLayout() );
      this.model = model;
      this.form = new Form();

      final JButton add = new JButton( (addButtonText == null
                                           ? "Add Row"
                                           : addButtonText),
                                       Images.getIcon( "/img/add.png" ) );
      add.setToolTipText( "Add another row for data entry" );
      add.addActionListener( new ActionListener()
      {
        @Override
        public void actionPerformed( final ActionEvent e )
        {
          model.addRow();
          processModel();
        }
      } );
      final JPanel p = new JPanel( new BorderLayout() );
      p.add( BorderLayout.WEST,
             add );
      p.setBorder( BorderFactory.createEmptyBorder( 0,
                                                    3,
                                                    12,
                                                    0 ) );

      add( BorderLayout.CENTER,
           form );
      add( BorderLayout.SOUTH,
           p );

      processModel();
    }


    private void processModel()
    {
      rows.clear();

      final int rowCount = model.rowCount();
      final int colCount = model.colCount();
      for( int r = 0; r < rowCount; r++ )
        {
          final List<Component> row = new ArrayList<>();
          rows.add( row );

          // Fill each column with a component from the Model
          for( int c = 0; c < colCount; c++ )
            {
              row.add( model.getComponent( r,
                                           c ) );
            }
        }
      rebuildLayout();
    }


    private void rebuildLayout()
    {
      form.clear();
      deleters.clear();

      int r = 0;
      for( final List<Component> row : rows )
        {
          int c = 0;
          for( final Component editor : row )
            {
              form.add( r,
                        c++,
                        editor,
                        1.0d,
                        1.0d,
                        1,
                        1,
                        0,
                        0,
                        Form.Sizing.WIDTH,
                        Form.Anchor.WEST );
            }
          // Add the 'del' button last
          final JButton del = new JButton( "Del",
                                           Images.getIcon( "/img/delete.png" ) );
          del.setToolTipText( "Delete this row (caution, there is no way to undo that!)" );
          deleters.put( del,
                        r );
          del.addActionListener( new ActionListener()
          {
            @Override
            public void actionPerformed( final ActionEvent e )
            {
              final int rowIndex = deleters.get( del );
              model.delRow( rowIndex );
              processModel();
            }
          } );
          form.add( r,
                    c,
                    del,
                    0.0d,
                    1.0d );
          r++;
        }

      revalidate();
      repaint();
    }


    public interface Model
    {
      public int colCount();


      public int rowCount();


      public Component getComponent( final int row,
                                     final int col );


      public void addRow();


      public void delRow( final int index );
    }

    private final Model model;
    private final Form form;
    private final List<List<Component>> rows = new ArrayList<>();
    private final Map<JButton,Integer> deleters = new HashMap<>();
    private static final long serialVersionUID = -1429878667549975065L;
  }


  public abstract class ComponentListener
    implements
      ActionListener,
      FocusListener,
      ItemListener
  {
    @Override
    public void actionPerformed( final ActionEvent e )
    {
      modified( e.getSource() );
    }


    @Override
    public void focusGained( final FocusEvent e )
    {
    }


    @Override
    public void focusLost( final FocusEvent e )
    {
      modified( e.getComponent() );
    }


    @Override
    public void itemStateChanged( final ItemEvent e )
    {
      modified( e.getSource() );
    }


    public abstract void modified( final Object source );
  }

  private static final long serialVersionUID = -869694310455034251L;
}
