package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.IOException;
import java.util.Properties;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPasswordField;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.ringlord.Filename;
import com.ringlord.boxer2.SigningProperties;
import com.ringlord.swing.Form;
import com.ringlord.swing.Images;


/**
 * <p>
 * A panel for editing the signing properties file contents without disturbing
 * other properties in the file. Recognized properties are:
 * </p>
 *
 * <ul>
 * <li>keystore-file — The filename of the Java Keystore
 * <li>keystore-pass — The password to access the given Java Keystore
 * <li>key-alias — The alias for the signing key in the Java Keystore
 * <li>key-pass — The password for the signing key in the Java Keystore
 * </ul>
 *
 * <p>
 * Two additional properties, if present, support the encryption of the
 * <tt>keystore-pass</tt> and <tt>key-pass</tt> values:
 * </p>
 *
 * <ul>
 * <li>encryption-info — Stores Algorithm, Base64-encoded Initialization Vector,
 * Base64-encoded Salt, and the constant PBKDF2 which denotes the key
 * strengthening algorithm used.
 * <li>encryption-checksum — The Base64-encoded SHA-256 checksum of the two
 * plain text passwords. After decrypting the passwords, this checksum is used
 * to determine whether the decryption was actually successful (i.e. the correct
 * password was used): The incorrect password would simply produce incorrect
 * decryption output, so this is our sanity check.
 * </ul>
 *
 * <p>
 * By storing this information in a file <em>external to the project</em>, it
 * becomes possible to include the Boxer configuration file (with a reference to
 * this properties file) as part of the project
 * <em>without exposing the keystore or passwords</em> to other developers, so
 * long as all developers agree on a location and name for this properties file.
 * Each developer can then sign the privately built JAR with his/her own key in
 * a Keystore that they built and maintain themselves.
 * </p>
 *
 * TODO Allow name of Signing Properties File to come from environment variable
 *
 * @author K Udo Schuermann
 */
public class SigningPropertiesEditor
  extends JPanel
{
  /**
   * Construct the editing panel for the given {@link File}, and place it within
   * the given {@link JDialog}.
   *
   * @param curFile
   *          The file whose contents to edit. The file need not exist; it will
   *          be created when the "Accept" button is pressed.
   * @param dialog
   *          The {@link JDialog} within which to edit the file; the dialog will
   *          be closed (made invisible) when the "Accept" or "Cancel" buttons
   *          are pressed or the close window button on the dialog itself is
   *          clicked.
   */
  public SigningPropertiesEditor( final File curFile,
                                  final JDialog dialog )
  {
    super( new BorderLayout() );

    setBorder( BorderFactory.createEmptyBorder( 8,
                                                12,
                                                8,
                                                12 ) );

    final Form editor = new Form( new Insets( 0,
                                              4,
                                              4,
                                              0 ) );
    final JButton pick;
    final JPanel filenamePick = new JPanel( new BorderLayout() );
    filenamePick.add( BorderLayout.CENTER,
                      keystoreFilename = new JTextField( 16 ) );
    filenamePick.add( BorderLayout.EAST,
                      pick = new JButton( "",
                                          Images.getIcon( "/img/folder.png" ) ) );
    pick.setToolTipText( "Pick a file" );
    editor.addRow( null,
                   encrypt = new JCheckBox( "Encrypt passwords in signing properties file",
                                            true ) );
    editor.addRow( algorithmLabel = new JLabel( "Cipher algorithm:" ),
                   algorithm = new JComboBox<>( cipherChoices ) );
    editor.addRow( pwd1Label = new JLabel( "Password:" ),
                   pwd1 = new JPasswordField( 16 ) );
    editor.addRow( pwd2Label = new JLabel( "Password (again):" ),
                   pwd2 = new JPasswordField( 16 ) );
    //
    editor.addRow( emptyLine = new JLabel( " " ) );
    //
    editor.addRow( new JLabel( "Keystore filename:" ),
                   filenamePick );
    editor.addRow( new JLabel( "Keystore password:" ),
                   keystorePassword = new JPasswordField( 16 ) );
    editor.addRow( new JLabel( "Signing key alias:" ),
                   keyAlias = new JTextField( 16 ) );
    editor.addRow( new JLabel( "Signing key password:" ),
                   keyPassword = new JPasswordField( 16 ) );

    encrypt.setToolTipText( "Check to encrypt the keystore password & key password" );
    algorithm.setToolTipText( "The encryption algorithm to use" );
    pwd1.setToolTipText( "The password to use for encrypting the keystore password & key password" );
    pwd2.setToolTipText( "The same password as above. To prevent painful mistakes, they must match" );

    keystoreFilename.setToolTipText( "The filename for the keystore file." );
    keystorePassword.setToolTipText( "<html>"
                                     + "The password for the keystore file.<br>"
                                     + "It will be encrypted if you choose<br>"
                                     + "using the options above." );
    keyAlias.setToolTipText( "The alias/name for the key in the keystore" );
    keyPassword.setToolTipText( "<html>"
                                + "The password for the signing key.<br>"
                                + "It will be encrypted if you choose<br>"
                                + "using the options above." );

    algorithm.setEditable( true );

    pick.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        boolean isFresh;
        if( fileChooser == null )
          {
            fileChooser = new JFileChooser();
            fileChooser.setDialogTitle( "Select Keystore File" );
            fileChooser.setFileSelectionMode( JFileChooser.FILES_ONLY );
            isFresh = true;
          }
        else
          {
            isFresh = false;
          }
        // final Settings config = Main.this.projectConfiguration.config();
        if( isFresh )
          {
            // this.fileChooser.setBounds( config.get(
            // "filechooser-window-xywh",
            // fileChooser.getBounds() ) );
          }
        fileChooser.setSelectedFile( Filename.resolveTilde( keystoreFilename.getText() ) );
        final int decision = fileChooser.showOpenDialog( SigningPropertiesEditor.this );
        if( decision == JFileChooser.APPROVE_OPTION )
          {
            final File f = fileChooser.getSelectedFile();
            keystoreFilename.setText( f.toString() );
          }
        // config.put( "filechooser-window-xywh",
        // fileChooser.getBounds() );
      }
      private JFileChooser fileChooser;
    } );

    final JPanel info = new JPanel( new GridLayout( 0,
                                                    1 ) );
    info.add( new JLabel( "This signing properties file should not be part of the project as it stores passwords" ) );
    info.add( new JLabel( "that would otherwise be accessible to all project members, especially if the file is" ) );
    info.add( new JLabel( "checked into a repository. The same goes for the keystore, too, as its certificates" ) );
    info.add( new JLabel( "and keys belong to and identify individuals!" ) );
    info.add( new JLabel() );
    info.add( new JLabel( "For an extra level of protection, use a password to encrypt the keystore passwords" ) );
    info.add( new JLabel( "in this signing properties file." ) );
    info.add( new JLabel() );
    info.add( new JLabel( "Four properties must be supplied for digital signing to function:" ) );
    info.add( new JLabel( "(1) Keystore filename, (2) Keystore password, (3) Key Alias, (4) Key Password." ) );

    info.setBorder( BorderFactory.createEmptyBorder( 4,
                                                     6,
                                                     24,
                                                     6 ) );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             info );
    top.add( BorderLayout.CENTER,
             editor );

    final JPanel grid = new JPanel( new GridLayout( 1,
                                                    0,
                                                    4,
                                                    0 ) );
    final JButton accept, cancel;
    grid.add( accept = new JButton( "Accept",
                                    Images.getIcon( "/img/accept.png" ) ) );
    grid.add( cancel = new JButton( "Cancel",
                                    Images.getIcon( "/img/cancel.png" ) ) );
    final JPanel cmd = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
    cmd.add( grid );
    cmd.setBorder( BorderFactory.createEmptyBorder( 8,
                                                    0,
                                                    0,
                                                    0 ) );

    add( BorderLayout.NORTH,
         top );
    add( BorderLayout.SOUTH,
         cmd );

    final SigningProperties signingInfo = load( curFile );
    if( signingInfo == null )
      {
        throw new IllegalStateException( "Unable to decrypt the file contents" );
      }

    accept.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        save( curFile,
              signingInfo );
        dialog.setVisible( false );
      }
    } );
    cancel.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        dialog.setVisible( false );
      }
    } );

    encrypt.addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent arg0 )
      {
        fixVisibility( editor,
                       accept );
      }
    } );

    final DocumentListener pwdCheck = new DocumentListener()
    {
      @Override
      public void changedUpdate( final DocumentEvent e )
      {
      }


      @Override
      public void insertUpdate( final DocumentEvent e )
      {
        checkPasswords( accept );
      }


      @Override
      public void removeUpdate( final DocumentEvent e )
      {
        checkPasswords( accept );
      }

    };

    pwd1.getDocument().addDocumentListener( pwdCheck );
    pwd2.getDocument().addDocumentListener( pwdCheck );

    checkPasswords( accept );

    fixVisibility( editor,
                   accept );
  }


  private void checkPasswords( final JButton accept )
  {
    final char[] p1 = pwd1.getPassword();
    final char[] p2 = pwd2.getPassword();
    if( new String( p1 ).equals( new String( p2 ) ) )
      {
        accept.setEnabled( true );
      }
    else
      {
        accept.setEnabled( false );
      }
  }


  private void fixVisibility( final Form form,
                              final JButton accept )
  {
    final boolean isEncrypted = encrypt.isSelected();
    algorithmLabel.setVisible( isEncrypted );
    algorithm.setVisible( isEncrypted );
    pwd1Label.setVisible( isEncrypted );
    pwd1.setVisible( isEncrypted );
    pwd2Label.setVisible( isEncrypted );
    pwd2.setVisible( isEncrypted );
    emptyLine.setVisible( isEncrypted );
    if( isEncrypted )
      {
        checkPasswords( accept );
      }
    else
      {
        accept.setEnabled( true );
      }
    form.revalidate();
    validate();
    repaint();
  }


  /**
   * Loads the {@link Properties} from the given file and initializes the GUI
   * components. The Properties object is kept by the caller, and when changes
   * are saved, it is updated, and passed to the {@link #save(File, Properties)}
   * method, thereby preserving any properties in the file that were there but
   * not used by this editor.
   *
   * @param file
   *          The file from which to load the Properties. No error is produced
   *          if the file does not exist.
   * @return The Properties object with all Properties found in the given file,
   *         or null if the file did not exist, or an IOException occurred for
   *         some reason while reading the file.
   */
  private SigningProperties load( final File file )
  {
    final SigningProperties ksInfo = SigningProperties.load( file,
                                                             new PasswordReaderGUI( this ) );
    if( ksInfo == null )
      {
        return null;
      }
    final boolean showCryptoFields = (ksInfo.getCipher() != null);

    encrypt.setSelected( showCryptoFields );
    algorithmLabel.setVisible( false );
    algorithm.setVisible( false );
    pwd1Label.setVisible( false );
    pwd1.setVisible( false );
    pwd2Label.setVisible( false );
    pwd2.setVisible( false );

    if( showCryptoFields )
      {
        algorithm.setSelectedItem( ksInfo.getCipher().getAlgorithm() );
        pwd1.setText( ksInfo.getPassword() );
        pwd2.setText( ksInfo.getPassword() );
      }
    else
      {
        algorithm.setSelectedIndex( 0 );
        pwd1.setText( "" );
        pwd2.setText( "" );
      }

    keystoreFilename.setText( ksInfo.getKeystoreFilename() );
    keystorePassword.setText( ksInfo.getKeystorePassword() );
    keyAlias.setText( ksInfo.getKeyAlias() );
    keyPassword.setText( ksInfo.getKeyPassword() );

    return ksInfo;
  }


  /**
   * Saves the {@link Properties} to the given file, using first a backup file.
   *
   * @param file
   *          The file to update.
   * @param signingInfo
   *          The Properties to save.
   */
  private void save( final File file,
                     final SigningProperties ksInfo )
  {
    ksInfo.setKeystoreFilename( Filename.resolveTilde( keystoreFilename.getText().trim() ).toString() );
    ksInfo.setKeystorePassword( new String( keystorePassword.getPassword() ) );
    ksInfo.setKeyAlias( keyAlias.getText().trim() );
    ksInfo.setKeyPassword( new String( keyPassword.getPassword() ) );

    try
      {
        if( encrypt.isSelected() && pwd1.getPassword().length > 0 )
          {
            ksInfo.save( file,
                         (String)algorithm.getSelectedItem(),
                         new String( pwd1.getPassword() ) );
          }
        else
          {
            ksInfo.save( file,
                         null,
                         null );
          }
      }
    catch( final IOException e )
      {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
  }

  private final JCheckBox encrypt;
  private final JComboBox<String> algorithm;
  private final JPasswordField pwd1, pwd2;
  private final JLabel algorithmLabel, pwd1Label, pwd2Label;
  private final JLabel emptyLine;
  //
  private final JTextField keystoreFilename, keyAlias;
  private final JPasswordField keystorePassword, keyPassword;
  //
  // NOTE: Due to the recent (2013) revelations that the United States National
  // Security Agency (NSA) has essentially compromised the security algorithms
  // and security mechanisms from many manufacturers of security hardware and
  // software (including Intel's hardware and perhaps Red Hat's Linux, too),
  // Boxer offers Bruce Schneier's Blowfish cipher as preferred over the US
  // Government's endorsed AES cipher (any endorsement by the US Government is
  // rather suspect at this stage, and reason to avoid it!)
  private static final String[] cipherChoices = new String[]{"Blowfish/CFB/NoPadding","AES/CBC/NoPadding"};
  private static final long serialVersionUID = 862696531982900654L;
}