package com.ringlord.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Desktop;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.URISyntaxException;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingWorker;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.text.Element;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.html.HTMLDocument;


/**
 * An "About" dialog, which consists of an optional logo image and HTML
 * <em>body</em> text (not a full document, only the portion between the
 * &lt;body&gt;…&lt;/body&gt; tags) with support for hyperlinks whose activation
 * launches the relevant URL using the system web browser. The HTML component
 * supports hover-highlighting of hyper links (coloring is currently hard
 * coded).
 *
 * @author K Udo Schueramnn
 */
public class About
  extends JPanel
{
  public About( final String htmlBody,
                final Font font,
                final Icon logo,
                final boolean offerJavaInfo )
  {
    super( new BorderLayout() );
    setBorder( BorderFactory.createEmptyBorder( 12,
                                                12,
                                                12,
                                                12 ) );

    final JPanel iconPane;
    final JButton java;
    if( (logo != null) || offerJavaInfo )
      {
        iconPane = new JPanel( new BorderLayout() );

        if( logo != null )
          {
            iconPane.add( BorderLayout.NORTH,
                          new JLabel( logo ) );
          }

        if( offerJavaInfo )
          {
            java = new JButton( "Java Info \u2026",
                                Images.getIcon( "/img/javabean.png" ) );
            java.setToolTipText( "<html>"
                                 + "Add Java environment information<br>"
                                 + "to the display field on the right." );
            final JPanel javaPane = new JPanel( new BorderLayout() );
            javaPane.add( BorderLayout.CENTER,
                          java );
            javaPane.setBorder( BorderFactory.createEmptyBorder( 0,
                                                                 0,
                                                                 2,
                                                                 0 ) );

            iconPane.add( BorderLayout.SOUTH,
                          javaPane );
            iconPane.setBorder( BorderFactory.createEmptyBorder( 0,
                                                                 0,
                                                                 0,
                                                                 12 ) );
          }
        else
          {
            java = null;
          }
      }
    else
      {
        iconPane = null;
        java = null;
      }

    final JEditorPane output = new JEditorPane();
    output.setFont( font );
    output.setContentType( "text/html" );
    output.setEditable( false );
    output.setBorder( BorderFactory.createEmptyBorder( 8,
                                                       8,
                                                       8,
                                                       8 ) );
    final HTMLDocument doc = (HTMLDocument)output.getDocument();
    final StyleContext ss = doc.getStyleSheet();
    final Style highlight = ss.addStyle( "HighlightLink",
                                         null );
    output.addHyperlinkListener( new HyperlinkListener()
    {
      @Override
      public void hyperlinkUpdate( final HyperlinkEvent e )
      {
        if( e.getEventType() == HyperlinkEvent.EventType.ENTERED )
          {
            // Highlight the link over which the mouse is hovering
            final Element anchor = e.getSourceElement();
            final int start = anchor.getStartOffset();
            final int end = anchor.getEndOffset();

            highlight.addAttribute( StyleConstants.Foreground,
                                    Color.RED );
            doc.setCharacterAttributes( start,
                                        end - start,
                                        highlight,
                                        false );
          }
        else if( e.getEventType() == HyperlinkEvent.EventType.EXITED )
          {
            // Unhighlight the previously highlighted link
            final Element anchor = e.getSourceElement();
            final int start = anchor.getStartOffset();
            final int end = anchor.getEndOffset();

            highlight.addAttribute( StyleConstants.Foreground,
                                    Color.BLUE );
            doc.setCharacterAttributes( start,
                                        end - start,
                                        highlight,
                                        false );
          }
        else if( e.getEventType() == HyperlinkEvent.EventType.ACTIVATED )
          {
            // Launch the system web browser with the clicked link
            final Desktop desktop = Desktop.getDesktop();
            new SwingWorker<Void,Void>()
            {
              @Override
              public Void doInBackground()
              {
                try
                  {
                    desktop.browse( e.getURL().toURI() );
                  }
                catch( final URISyntaxException x )
                  {
                    x.printStackTrace( System.err );
                  }
                catch( final IOException x )
                  {
                    x.printStackTrace( System.err );
                  }
                return null;
              }
            }.execute();
          }
      }
    } );

    if( java != null )
      {
        java.addActionListener( new ActionListener()
        {
          @Override
          public void actionPerformed( final ActionEvent e )
          {
            final StringBuilder sb = new StringBuilder( "<br><hr>" );
            final Runtime rt = Runtime.getRuntime();

            sb.append( "Java " )
              .append( System.getProperty( "java.runtime.version" ) )
              .append( " (" )
              .append( System.getProperty( "java.vm.name" ) )
              .append( " " )
              .append( System.getProperty( "java.vm.version" ) )
              .append( ")<br>" )
              .append( System.getProperty( "os.name" ) )
              .append( " " )
              .append( System.getProperty( "os.version" ) )
              .append( " (" )
              .append( System.getProperty( "os.arch" ) )
              .append( ")" )
              .append( "<br>Available processors: " )
              .append( rt.availableProcessors() )
              .append( "<br>Memory: " )
              .append( String.format( "%,1.2fM of %,1.2fM",
                                      rt.freeMemory() / (1024.0d * 1024.0d),
                                      rt.maxMemory() / (1024.0d * 1024.0d) ) )
              .append( "<br>Country_language: " )
              .append( System.getProperty( "user.country" ) )
              .append( "_" )
              .append( System.getProperty( "user.language" ) )
              .append( "<br>Default file encoding: " )
              .append( System.getProperty( "file.encoding" ) )
              .append( "<br>Graphics environment: " )
              .append( System.getProperty( "java.awt.graphicsenv" ) );

            output.setText( aboutHead + htmlBody + sb + aboutTail );
          }
        } );
      }

    final JScrollPane sOutput = new JScrollPane( output );

    if( iconPane != null )
      {
        add( BorderLayout.WEST,
             iconPane );
      }
    add( BorderLayout.CENTER,
         sOutput );

    output.setText( aboutHead + htmlBody + aboutTail );
  }
  private static final long serialVersionUID = 7294034656545407785L;
  private static final String aboutHead = "<!doctype html><html>"
                                          + "<head><style>body{font-size:100%;}</style></head><body>";
  private static final String aboutTail = "</body>" + "</html>";
}
