package com.ringlord.swing;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;
import javax.swing.MutableComboBoxModel;
import javax.swing.border.Border;
import javax.swing.event.DocumentListener;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import javax.swing.text.JTextComponent;


/**
 * <p>
 * An editable JComboBox for String items. Typing items that are not yet in the
 * list (and pressing ENTER) will add them. Removing all characters from a
 * selected item (and pressing ENTER) removes the item again.
 * </p>
 *
 * @author K Udo Schuermann
 **/
public class ComboText
  extends JComboBox<String>
{
  private static final long serialVersionUID = -1435525280466543560L;


  public ComboText( final MutableComboBoxModel<String> model,
                    final int maximumRemembered )
  {
    super( model );
    this.model = model;
    this.editor = getEditor();
    this.maximumRemembered = maximumRemembered;

    model.insertElementAt( "",
                           curIndex = 0 );

    editor.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        editorAction();
      }
    } );
    editor.getEditorComponent().addFocusListener( new FocusListener()
    {
      @Override
      public void focusGained( final FocusEvent e )
      {
      }


      @Override
      public void focusLost( final FocusEvent e )
      {
        editorAction();
      }
    } );
    addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        if( pickingFromList )
          {
            final String item = (String)getSelectedItem();
            final int size = model.getSize();
            for( int i = 0; i < size; i++ )
              {
                if( item.equals( model.getElementAt( i ) ) )
                  {
                    if( PROMOTE_SELECTION )
                      {
                        if( i > 1 )
                          {
                            // move the element to the top-spot
                            model.removeElementAt( i );
                            model.insertElementAt( item,
                                                   curIndex = 1 );
                            setSelectedItem( item );
                          }
                      }
                    else
                      {
                        curIndex = i;
                      }
                    break;
                  }
              }
          }
        pickingFromList = false;
      }
    } );
    addPopupMenuListener( new PopupMenuListener()
    {
      @Override
      public void popupMenuCanceled( final PopupMenuEvent e )
      {
      }


      @Override
      public void popupMenuWillBecomeVisible( final PopupMenuEvent e )
      {
        pickingFromList = true;
      }


      @Override
      public void popupMenuWillBecomeInvisible( final PopupMenuEvent e )
      {
      }
    } );
  }


  public void addDocumentListener( final DocumentListener l )
  {
    final Component c = getEditor().getEditorComponent();
    if( c instanceof JTextComponent )
      {
        ((JTextComponent)c).getDocument().addDocumentListener( l );
      }
  }


  public void removeDocumentListener( final DocumentListener l )
  {
    final Component c = getEditor().getEditorComponent();
    if( c instanceof JTextComponent )
      {
        ((JTextComponent)c).getDocument().removeDocumentListener( l );
      }
  }


  @Override
  public void addFocusListener( final FocusListener l )
  {
    final Component c = getEditor().getEditorComponent();
    if( c instanceof JTextComponent )
      {
        ((JTextComponent)c).addFocusListener( l );
      }
  }


  @Override
  public void removeFocusListener( final FocusListener l )
  {
    final Component c = getEditor().getEditorComponent();
    if( c instanceof JTextComponent )
      {
        ((JTextComponent)c).removeFocusListener( l );
      }
  }


  public String getText()
  {
    final Component c = getEditor().getEditorComponent();
    if( c instanceof JTextComponent )
      {
        return ((JTextComponent)c).getText();
      }
    return (String)getSelectedItem();
  }


  public void finishEditing()
  {
    editorAction();
  }


  @Override
  public void setSelectedIndex( final int index )
  {
    curIndex = index;
    super.setSelectedIndex( index );
  }


  private void editorAction()
  {
    add( (String)editor.getItem() );
  }


  public void add( final List<String> items )
  {
    for( int i = items.size() - 1; i >= 0; i-- )
      {
        add( items.get( i ) );
      }
  }


  public void add( final String item )
  {
    if( "".equals( item ) )
      {
        // user blanked out a previously existing item, so we'll delete it from
        // the list, but do not remove the first (blank) item.
        if( (curIndex > 0) && (curIndex < model.getSize()) )
          {
            model.removeElementAt( curIndex );
          }
        setSelectedItem( "" ); // pick the first (blank) item
        curIndex = 0;
      }
    else
      // a non-blank item
      {
        final int size = model.getSize();
        for( int i = 0; i < size; i++ )
          {
            if( item.equals( model.getElementAt( i ) ) )
              {
                if( PROMOTE_SELECTION )
                  {
                    if( i > 1 )
                      {
                        // Existing entry, move it to the top-spot
                        model.removeElementAt( i );
                        model.insertElementAt( item,
                                               curIndex = 1 );
                        setSelectedItem( item );
                      }
                  }
                else
                  {
                    curIndex = i;
                  }
                return;
              }
          }
        // It's a new entry
        model.insertElementAt( item,
                               curIndex = 1 );
        setSelectedItem( item );
      }
  }


  public class Renderer
    extends JLabel
    implements
      ListCellRenderer<String>
  {
    private static final long serialVersionUID = -9031887398642963679L;


    public Renderer()
    {
      setOpaque( true );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends String> list,
                                                   final String value,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      // NOTE: The JComboBox appears not to honor our colors and border... :-/
      setText( value.toString() );
      if( isSelected )
        {
          list.setForeground( list.getSelectionForeground() );
          list.setBackground( list.getSelectionBackground() );
        }
      else
        {
          list.setForeground( Color.black );
          list.setBackground( index > maximumRemembered
              ? warnColor
              : happyColor );
        }
      setBorder( hasFocus
          ? border1
          : border0 );
      return this;
    }
  }

  private boolean pickingFromList;
  private int curIndex;
  //
  private final int maximumRemembered;
  private final MutableComboBoxModel<String> model;
  //
  private static final Color happyColor = new Color( 168,
                                                     232,
                                                     198 );
  private static final Color warnColor = new Color( 222,
                                                    168,
                                                    168 );
  private static final Border border0 = BorderFactory.createEmptyBorder( 1,
                                                                         1,
                                                                         1,
                                                                         1 );
  private static final Border border1 = BorderFactory.createLineBorder( Color.black,
                                                                        1 );
  private static final boolean PROMOTE_SELECTION = true;

}
